"""
Module for specialized parameters. The :mod:`qcodes.instrument.parameter`
module provides generic parameters for different generic cases. This module
provides useful/convenient specializations of such generic parameters.
"""

from time import perf_counter
from typing import Any

from qcodes.instrument.parameter import Parameter


class ElapsedTimeParameter(Parameter):
    """
    Parameter to measure elapsed time. Measures wall clock time since the
    last reset of the instance's clock. The clock is reset upon creation of the
    instance. The constructor passes kwargs along to the Parameter constructor.

    Args:
        name: The local name of the parameter. See the documentation of
            :class:`qcodes.instrument.parameter.Parameter` for more details.
    """

    def __init__(self, name: str, label: str = 'Elapsed time', **kwargs: Any):

        hardcoded_kwargs = ['unit', 'get_cmd', 'set_cmd']

        for hck in hardcoded_kwargs:
            if hck in kwargs:
                raise ValueError(f'Can not set "{hck}" for an '
                                 'ElapsedTimeParameter.')

        super().__init__(name=name,
                         label=label,
                         unit='s',
                         set_cmd=False,
                         **kwargs)

        self._t0: float = perf_counter()

    def get_raw(self) -> float:
        return perf_counter() - self.t0

    def reset_clock(self) -> None:
        self._t0 = perf_counter()

    @property
    def t0(self) -> float:
        return self._t0



from typing import Union, Dict, Sequence, List, Optional

import time
import re


import numpy as np
import qcodes as qc


_guid_pattern = re.compile(r'^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$')


def generate_guid(timeint: Union[int, None]=None,
                  sampleint: Union[int, None]=None) -> str:
    """
    Generate a guid string to go into the GUID column of the runs table.
    The GUID is based on the GUID-components in the qcodesrc file.
    The generated string is of the format
    '12345678-1234-1234-1234-123456789abc', where the first eight hex numbers
    comprise the 4 byte sample code, the next 2 hex numbers comprise the 1 byte
    location, the next 2+4 hex numbers are the 3 byte work station code, and
    the final 4+12 hex number give the 8 byte integer time in ms since epoch
    time

    Args:
        timeint: An integer of miliseconds since unix epoch time
        sampleint: A code for the sample
    """
    cfg = qc.config

    try:
        guid_comp = cfg['GUID_components']
    except KeyError:
        raise RuntimeError('Invalid QCoDeS config file! No GUID_components '
                           'specified. Can not proceed.')

    location = guid_comp['location']
    station = guid_comp['work_station']

    if timeint is None:
        # ms resolution, checked on windows
        timeint = int(np.round(time.time()*1000))
    if sampleint is None:
        sampleint = guid_comp['sample']

    if sampleint == 0:
        sampleint = int('a'*8, base=16)

    loc_str = f'{location:02x}'
    stat_str = f'{station:06x}'
    smpl_str = f'{sampleint:08x}'
    time_str = f'{timeint:016x}'

    guid = (f'{smpl_str}-{loc_str}{stat_str[:2]}-{stat_str[2:]}-'
            f'{time_str[:4]}-{time_str[4:]}')

    return guid


def parse_guid(guid: str) -> Dict[str, int]:
    """
    Parse a guid back to its four constituents

    Args:
        guid: a valid guid str

    Returns:
        A dict with keys 'location', 'work_station', 'sample', and 'time'
          as integer values
    """
    guid = guid.replace('-', '')
    components = {}
    components['sample'] = int(guid[:8], base=16)
    components['location'] = int(guid[8:10], base=16)
    components['work_station'] = int(guid[10:16], base=16)
    components['time'] = int(guid[16:], base=16)

    return components


def set_guid_location_code() -> None:
    """
    Interactive function to set the location code.
    """
    cfg = qc.config
    old_loc = cfg['GUID_components']['location']
    print(f'Updating GUID location code. Current location code is: {old_loc}')
    if old_loc != 0:
        print('That is a non-default location code. Perhaps you should not '
              'change it? Re-enter that code to leave it unchanged.')
    loc_str = input('Please enter the new location code (1-256): ')
    try:
        location = int(loc_str)
    except ValueError:
        print('The location code must be an integer. No update performed.')
        return
    if not(257 > location > 0):
        print('The location code must be between 1 and 256 (both included). '
              'No update performed')
        return

    cfg['GUID_components']['location'] = location
    cfg.save_to_home()


def set_guid_work_station_code() -> None:
    """
    Interactive function to set the work station code
    """
    cfg = qc.config
    old_ws = cfg['GUID_components']['work_station']
    print('Updating GUID work station code. '
          f'Current work station code is: {old_ws}')
    if old_ws != 0:
        print('That is a non-default work station code. Perhaps you should not'
              ' change it? Re-enter that code to leave it unchanged.')
    ws_str = input('Please enter the new work station code (1-16777216): ')
    try:
        work_station = int(ws_str)
    except ValueError:
        print('The work station code must be an integer. No update performed.')
        return
    if not(16777216 > work_station > 0):
        print('The work staion code must be between 1 and 256 (both included).'
              ' No update performed')
        return

    cfg['GUID_components']['work_station'] = work_station
    cfg.save_to_home()


def filter_guids_by_parts(guids: Sequence[str],
                          location: Optional[int] = None,
                          sample_id: Optional[int] = None,
                          work_station: Optional[int] = None) -> List[str]:
    """
    Filter a sequence of GUIDs by location, sample_id and/or work_station.

    Args:
        guids: Sequence of guids that should be filtered.
        location: Location code to match
        sample_id: Sample_id to match
        work_station: Workstation to match

    Returns:
        A list of GUIDs that matches the supplied parts.
    """
    matched_guids = []
    for guid in guids:
        guid_dict = parse_guid(guid)
        match = True
        if sample_id is not None:
            if guid_dict['sample'] != sample_id:
                match = False
        if location is not None:
            if guid_dict['location'] != location:
                match = False
        if work_station is not None:
            if guid_dict['work_station'] != work_station:
                match = False

        if match:
            matched_guids.append(guid)
    return matched_guids


def validate_guid_format(guid: str) -> None:
    """
    Validate the format of the given guid. This function does not check the
    correctness of the data inside the guid (e.g. timestamps in the far
    future)
    """
    if _guid_pattern.match(guid):
            return
    else:
        raise ValueError(f'Did not receive a valid guid. Got {guid}')



import pytest

from qcodes.utils.command import Command, NoCommandError


class CustomError(Exception):
    pass


def test_bad_calls():
    with pytest.raises(TypeError):
        Command()

    with pytest.raises(TypeError):
        Command(cmd='')

    with pytest.raises(TypeError):
        Command(0, '', output_parser=lambda: 1)

    with pytest.raises(TypeError):
        Command(1, '', input_parser=lambda: 1)

    with pytest.raises(TypeError):
        Command(0, cmd='', exec_str='not a function')

    with pytest.raises(TypeError):
        Command(0, cmd=lambda: 1, no_cmd_function='not a function')


def test_no_cmd():
    with pytest.raises(NoCommandError):
        Command(0)

    def no_cmd_function():
        raise CustomError('no command')

    no_cmd = Command(0, no_cmd_function=no_cmd_function)
    with pytest.raises(CustomError):
        no_cmd()


def test_cmd_str():
    def f_now(x):
        return x + ' now'

    def upper(s):
        return s.upper()

    def reversestr(s):
        return s[::-1]

    def swap(a, b):
        return b, a

    # basic exec_str
    cmd = Command(0, 'pickles', exec_str=f_now)
    assert cmd() == 'pickles now'

    # with output parsing
    cmd = Command(0, 'blue', exec_str=f_now, output_parser=upper)
    assert cmd() == 'BLUE NOW'

    # parameter insertion
    cmd = Command(3, '{} is {:.2f}% better than {}', exec_str=f_now)
    assert cmd('ice cream', 56.2, 'cake') == \
                     'ice cream is 56.20% better than cake now'
    with pytest.raises(ValueError):
        cmd('cake', 'a whole lot', 'pie')

    with pytest.raises(TypeError):
        cmd('donuts', 100, 'bagels', 'with cream cheese')

    # input parsing
    cmd = Command(1, 'eat some {}', exec_str=f_now, input_parser=upper)
    assert cmd('ice cream') == 'eat some ICE CREAM now'

    # input *and* output parsing
    cmd = Command(1, 'eat some {}', exec_str=f_now,
                  input_parser=upper, output_parser=reversestr)
    assert cmd('ice cream') == 'won MAERC ECI emos tae'

    # multi-input parsing, no output parsing
    cmd = Command(2, '{} and {}', exec_str=f_now, input_parser=swap)
    assert cmd('I', 'you') == 'you and I now'

    # multi-input parsing *and* output parsing
    cmd = Command(2, '{} and {}', exec_str=f_now,
                  input_parser=swap, output_parser=upper)
    assert cmd('I', 'you') == 'YOU AND I NOW'


def test_cmd_function():
    def myexp(a, b):
        return a ** b

    cmd = Command(2, myexp)
    assert cmd(10, 3) == 1000

    with pytest.raises(TypeError):
        Command(3, myexp)

    # with output parsing
    cmd = Command(2, myexp, output_parser=lambda x: 5 * x)
    assert cmd(10, 3) == 5000

    # input parsing
    cmd = Command(1, abs, input_parser=lambda x: x + 1)
    assert cmd(-10) == 9

    # input *and* output parsing
    cmd = Command(1, abs, input_parser=lambda x: x + 2,
                  output_parser=lambda x: 3 * x)
    assert cmd(-6) == 12

    # multi-input parsing, no output parsing
    cmd = Command(2, myexp, input_parser=lambda x, y: (y, x))
    assert cmd(3, 10) == 1000

    # multi-input parsing *and* output parsing
    cmd = Command(2, myexp, input_parser=lambda x, y: (y, x),
                  output_parser=lambda x: 10 * x)
    assert cmd(8, 2) == 2560



# awg file -> (what, we, put, into, make_send_and_load_awg_file)
# This module parses an awg file using THREE sub-parser. This code could
# probably be streamlined somewhat.
import struct
from typing import Any, Dict, List, Tuple, Union

import numpy as np

AWG_FILE_FORMAT = {
    'MAGIC': 'h',
    'VERSION': 'h',
    'SAMPLING_RATE': 'd',    # d
    'REPETITION_RATE': 'd',    # # NAME?
    'HOLD_REPETITION_RATE': 'h',    # True | False
    'CLOCK_SOURCE': 'h',    # Internal | External
    'REFERENCE_SOURCE': 'h',    # Internal | External
    'EXTERNAL_REFERENCE_TYPE': 'h',    # Fixed | Variable
    'REFERENCE_CLOCK_FREQUENCY_SELECTION': 'h',  # 10 MHz | 20 MHz | 100 MHz
    'REFERENCE_MULTIPLIER_RATE': 'h',    #
    'DIVIDER_RATE': 'h',   # 1 | 2 | 4 | 8 | 16 | 32 | 64 | 128 | 256
    'TRIGGER_SOURCE': 'h',    # Internal | External
    'INTERNAL_TRIGGER_RATE': 'd',    #
    'TRIGGER_INPUT_IMPEDANCE': 'h',    # 50 ohm | 1 kohm
    'TRIGGER_INPUT_SLOPE': 'h',    # Positive | Negative
    'TRIGGER_INPUT_POLARITY': 'h',    # Positive | Negative
    'TRIGGER_INPUT_THRESHOLD': 'd',    #
    'EVENT_INPUT_IMPEDANCE': 'h',    # 50 ohm | 1 kohm
    'EVENT_INPUT_POLARITY': 'h',    # Positive | Negative
    'EVENT_INPUT_THRESHOLD': 'd',
    'JUMP_TIMING': 'h',    # Sync | Async
    'INTERLEAVE': 'h',    # On |  This setting is stronger than coupling
    'ZEROING': 'h',    # On | Off
    'COUPLING': 'h',    # The Off | Pair | All setting is weaker than .
    'RUN_MODE': 'h',    # Continuous | Triggered | Gated | Sequence
    'WAIT_VALUE': 'h',    # First | Last
    'RUN_STATE': 'h',    # On | Off
    'INTERLEAVE_ADJ_PHASE': 'd',
    'INTERLEAVE_ADJ_AMPLITUDE': 'd',
    'EVENT_JUMP_MODE': 'h',  # Event jump | Dynamic jump
    'TABLE_JUMP_STROBE': 'h',  # On
    'TABLE_JUMP_DEFINITION': 'ignore',  # Array of tablejump
    # ---------------------
    # Channel 1 settings
    # ---------------------
    'OUTPUT_WAVEFORM_NAME_1': 's',  # if in continuous mode
    'DAC_RESOLUTION_1': 'h',  # 8 | 10
    'CHANNEL_STATE_1': 'h',  # On | Off
    'ANALOG_DIRECT_OUTPUT_1': 'h',  # On | Off
    'ANALOG_FILTER_1': 'h',  # Enum type.
    'ANALOG_METHOD_1': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'ANALOG_AMPLITUDE_1': 'd',
    # When the Input Method is High/Low, it is skipped.
    'ANALOG_OFFSET_1': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'ANALOG_HIGH_1': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'ANALOG_LOW_1': 'd',
    'MARKER1_SKEW_1': 'd',
    'MARKER1_METHOD_1': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'MARKER1_AMPLITUDE_1': 'd',
    # When the Input Method is High/Low, it is skipped.
    'MARKER1_OFFSET_1': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER1_HIGH_1': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER1_LOW_1': 'd',
    'MARKER2_SKEW_1': 'd',
    'MARKER2_METHOD_1': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'MARKER2_AMPLITUDE_1': 'd',
    # When the Input Method is High/Low, it is skipped.
    'MARKER2_OFFSET_1': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER2_HIGH_1': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER2_LOW_1': 'd',
    'DIGITAL_METHOD_1': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'DIGITAL_AMPLITUDE_1': 'd',
    # When the Input Method is High/Low, it is skipped.
    'DIGITAL_OFFSET_1': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'DIGITAL_HIGH_1': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'DIGITAL_LOW_1': 'd',
    'EXTERNAL_ADD_1': 'h',  # AWG5000 only
    'PHASE_DELAY_INPUT_METHOD_1':   'h',  # Phase/DelayInme/DelayInints
    'PHASE_1': 'd',  # When the Input Method is not Phase, it is skipped.
    # When the Input Method is not DelayInTime, it is skipped.
    'DELAY_IN_TIME_1': 'd',
    # When the Input Method is not DelayInPoint, it is skipped.
    'DELAY_IN_POINTS_1': 'd',
    'CHANNEL_SKEW_1': 'd',
    'DC_OUTPUT_LEVEL_1': 'd',  # V
    # ---------------------
    # ---------------------
    'OUTPUT_WAVEFORM_NAME_2': 's',
    'DAC_RESOLUTION_2': 'h',  # 8 | 10
    'CHANNEL_STATE_2': 'h',  # On | Off
    'ANALOG_DIRECT_OUTPUT_2': 'h',  # On | Off
    'ANALOG_FILTER_2': 'h',  # Enum type.
    'ANALOG_METHOD_2': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'ANALOG_AMPLITUDE_2': 'd',
    # When the Input Method is High/Low, it is skipped.
    'ANALOG_OFFSET_2': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'ANALOG_HIGH_2': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'ANALOG_LOW_2': 'd',
    'MARKER1_SKEW_2': 'd',
    'MARKER1_METHOD_2': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'MARKER1_AMPLITUDE_2': 'd',
    # When the Input Method is High/Low, it is skipped.
    'MARKER1_OFFSET_2': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER1_HIGH_2': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER1_LOW_2': 'd',
    'MARKER2_SKEW_2': 'd',
    'MARKER2_METHOD_2': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'MARKER2_AMPLITUDE_2': 'd',
    # When the Input Method is High/Low, it is skipped.
    'MARKER2_OFFSET_2': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER2_HIGH_2': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER2_LOW_2': 'd',
    'DIGITAL_METHOD_2': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'DIGITAL_AMPLITUDE_2': 'd',
    # When the Input Method is High/Low, it is skipped.
    'DIGITAL_OFFSET_2': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'DIGITAL_HIGH_2': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'DIGITAL_LOW_2': 'd',
    'EXTERNAL_ADD_2': 'h',  # AWG5000 only
    'PHASE_DELAY_INPUT_METHOD_2':   'h',  # Phase/DelayInme/DelayInints
    'PHASE_2': 'd',  # When the Input Method is not Phase, it is skipped.
    # When the Input Method is not DelayInTime, it is skipped.
    'DELAY_IN_TIME_2': 'd',
    # When the Input Method is not DelayInPoint, it is skipped.
    'DELAY_IN_POINTS_2': 'd',
    'CHANNEL_SKEW_2': 'd',
    'DC_OUTPUT_LEVEL_2': 'd',  # V
    # ---------------------
    # ---------------------
    'OUTPUT_WAVEFORM_NAME_3': 's',
    'DAC_RESOLUTION_3': 'h',  # 8 | 10
    'CHANNEL_STATE_3': 'h',  # On | Off
    'ANALOG_DIRECT_OUTPUT_3': 'h',  # On | Off
    'ANALOG_FILTER_3': 'h',  # Enum type.
    'ANALOG_METHOD_3': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'ANALOG_AMPLITUDE_3': 'd',
    # When the Input Method is High/Low, it is skipped.
    'ANALOG_OFFSET_3': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'ANALOG_HIGH_3': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'ANALOG_LOW_3': 'd',
    'MARKER1_SKEW_3': 'd',
    'MARKER1_METHOD_3': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'MARKER1_AMPLITUDE_3': 'd',
    # When the Input Method is High/Low, it is skipped.
    'MARKER1_OFFSET_3': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER1_HIGH_3': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER1_LOW_3': 'd',
    'MARKER2_SKEW_3': 'd',
    'MARKER2_METHOD_3': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'MARKER2_AMPLITUDE_3': 'd',
    # When the Input Method is High/Low, it is skipped.
    'MARKER2_OFFSET_3': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER2_HIGH_3': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER2_LOW_3': 'd',
    'DIGITAL_METHOD_3': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'DIGITAL_AMPLITUDE_3': 'd',
    # When the Input Method is High/Low, it is skipped.
    'DIGITAL_OFFSET_3': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'DIGITAL_HIGH_3': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'DIGITAL_LOW_3': 'd',
    'EXTERNAL_ADD_3': 'h',  # AWG5000 only
    'PHASE_DELAY_INPUT_METHOD_3':   'h',  # Phase/DelayInme/DelayInints
    'PHASE_3': 'd',  # When the Input Method is not Phase, it is skipped.
    # When the Input Method is not DelayInTime, it is skipped.
    'DELAY_IN_TIME_3': 'd',
    # When the Input Method is not DelayInPoint, it is skipped.
    'DELAY_IN_POINTS_3': 'd',
    'CHANNEL_SKEW_3': 'd',
    'DC_OUTPUT_LEVEL_3': 'd',  # V
    # ---------------------
    # ---------------------
    'OUTPUT_WAVEFORM_NAME_4': 's',
    'DAC_RESOLUTION_4': 'h',  # 8 | 10
    'CHANNEL_STATE_4': 'h',  # On | Off
    'ANALOG_DIRECT_OUTPUT_4': 'h',  # On | Off
    'ANALOG_FILTER_4': 'h',  # Enum type.
    'ANALOG_METHOD_4': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'ANALOG_AMPLITUDE_4': 'd',
    # When the Input Method is High/Low, it is skipped.
    'ANALOG_OFFSET_4': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'ANALOG_HIGH_4': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'ANALOG_LOW_4': 'd',
    'MARKER1_SKEW_4': 'd',
    'MARKER1_METHOD_4': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'MARKER1_AMPLITUDE_4': 'd',
    # When the Input Method is High/Low, it is skipped.
    'MARKER1_OFFSET_4': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER1_HIGH_4': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER1_LOW_4': 'd',
    'MARKER2_SKEW_4': 'd',
    'MARKER2_METHOD_4': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'MARKER2_AMPLITUDE_4': 'd',
    # When the Input Method is High/Low, it is skipped.
    'MARKER2_OFFSET_4': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER2_HIGH_4': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'MARKER2_LOW_4': 'd',
    'DIGITAL_METHOD_4': 'h',  # Amplitude/Offset, High/Low
    # When the Input Method is High/Low, it is skipped.
    'DIGITAL_AMPLITUDE_4': 'd',
    # When the Input Method is High/Low, it is skipped.
    'DIGITAL_OFFSET_4': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'DIGITAL_HIGH_4': 'd',
    # When the Input Method is Amplitude/Offset, it is skipped.
    'DIGITAL_LOW_4': 'd',
    'EXTERNAL_ADD_4': 'h',  # AWG5000 only
    'PHASE_DELAY_INPUT_METHOD_4':   'h',  # Phase/DelayInme/DelayInints
    'PHASE_4': 'd',  # When the Input Method is not Phase, it is skipped.
    # When the Input Method is not DelayInTime, it is skipped.
    'DELAY_IN_TIME_4': 'd',
    # When the Input Method is not DelayInPoint, it is skipped.
    'DELAY_IN_POINTS_4': 'd',
    'CHANNEL_SKEW_4': 'd',
    'DC_OUTPUT_LEVEL_4': 'd',  # V
    }

# Note: this dictionary is dynamically updated by _parser1
AWG_FILE_FORMAT_WAV = {
    'WAVEFORM_NAME': 's',
    'WAVEFORM_TYPE': 'h',
    'WAVEFORM_LENGTH': 'l',
    'WAVEFORM_TIMESTAMP': '8H',
    'WAVEFORM_DATA_4': None
    }

AWG_FILE_FORMAT_SEQ = {
    'SEQUENCE_WAIT': 'h',  # On | Off
    'SEQUENCE_LOOP': 'l',  # 0=infinite
    'SEQUENCE_JUMP': 'h',  # OFF:0, INDEX: element #, NEXT: -1
    'SEQUENCE_GOTO': 'h',  # 0 if GOTO is OFF, else an element number
    'SEQUENCE_WAVEFORM': 's',
    'SEQUENCE_IS': '2h',  # Yes | No
    'SEQUENCE_SUBSEQ': 's'
    }

AWG_TRANSLATER = {
    'HOLD_REPETITION_RATE': {0: 'False', 1: 'True'},
    'CLOCK_SOURCE': {1: 'Internal', 2: 'External'},
    'REFERENCE_SOURCE': {1: 'Internal', 2: 'External'},
    'EXTERNAL_REFERENCE_TYPE': {1: 'Fixed', 2: 'Variable'},
    'TRIGGER_SOURCE': {1: 'Internal', 2: 'External'},
    'TRIGGER_INPUT_IMPEDANCE': {1: '50 Ohm', 2: '1 kOhm'},
    'TRIGGER_INPUT_SLOPE': {1: 'Positive', 2: 'Negative'},
    'TRIGGER_INPUT_POLARITY': {1: 'Positive', 2: 'Negative'},
    'EVENT_INPUT_IMPEDANCE': {1: '50 Ohm', 2: '1 kOhm'},
    'EVENT_INPUT_SLOPE': {1: 'Positive', 2: 'Negative'},
    'EVENT_INPUT_POLARITY': {1: 'Positive', 2: 'Negative'},
    'JUMP_TIMING': {1: 'Sync', 2: 'Async'},
    'RUN_MODE': {1: 'Continuous', 2: 'Triggered', 3: 'Gated', 4: 'Sequence'},
    'WAIT_VALUE': {1: 'First', 2: 'Last'}
    }

_parser3_output = Tuple[
    List[List[Dict[Any, Any]]],
    List[List[Dict[Any, Any]]],
    List[List[Dict[Any, Any]]],
    List[Union[str, int]],
    List[Union[str, int]],
    List[Union[str, int]],
    List[Union[str, int]],
    List[int]
]


def _unpacker(
        binaryarray: np.ndarray,
        dacbitdepth: int = 14
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Unpacks an awg-file integer wave into a waveform and two markers
    in the same way as the AWG does. This can be useful for checking
    how the signals are going to be interpreted by the instrument.

    Args:
        binaryarray: A numpy array containing the
            packed waveform and markers.
        dacbitdepth: Specifies the bit depth for the digitisation
            of the waveform. Allowed values: 14, 8. Default: 14.

    Returns:
        The waveform scaled to have values from -1 to 1, marker 1, marker 2.
    """

    wflength = len(binaryarray)
    wf = np.zeros(wflength)
    m1 = np.zeros(wflength)
    m2 = np.zeros(wflength)

    for ii, bitnum in enumerate(binaryarray):
        bitstring = bin(bitnum)[2:].zfill(16)
        m2[ii] = int(bitstring[0])
        m1[ii] = int(bitstring[1])
        wf[ii] = (int(bitstring[2:], base=2)-2**13)/2**13
        # print(bitstring, int(bitstring[2:], base=2))

    return wf, m1, m2


def _unwrap(bites: bytes, fmt: str) -> Union[str, int, Tuple[Any, ...]]:
    """
    Helper function for interpreting the bytes from the awg file.

    Args:
        bites: a bytes object
        fmt: the format string (either 's', 'h' or 'd')

    """
    value: Union[str, int, Tuple[Any, ...]]
    if fmt == 's':
        value = bites[:-1].decode('ascii')
    elif fmt == 'ignore':
        value = 'Not read'
    else:
        value = struct.unpack('<'+fmt, bites)
        if len(value) == 1:
            value = value[0]

    return value


def _getendingnumber(string: str) -> Tuple[int, str]:
    """
    Helper function to extract the last number of a string

    Args:
        string: A .awg field name, like SEQUENCE_JUMP_23

    Returns:
        The number and the shortened string,
        e.g. 'SEQUENCE_JUMP_23' -> (23, 'SEQUENCE_JUMP_')
    """

    num = ''

    for char in string[::-1]:
        if char.isdigit():
            num += char
        else:
            break

    return int(num[::-1]), string[:-len(num)]


awgfilepath = ('/Users/william/AuxiliaryQCoDeS/AWGhelpers/awgfiles/' +
               'customawgfile.awg')

awgfilepath2 = ('/Users/william/AuxiliaryQCoDeS/AWGhelpers/awgfiles/' +
                'machinemadefortest.awg')


def _parser1(
        awgfilepath: str
) -> Tuple[Dict[str, Union[str, int, Tuple[Any, ...]]], List[List[Any]], List[List[Any]]]:
    """
    Helper function doing the heavy lifting of reading and understanding the
    binary .awg file format.

    Args:
        awgfilepath: The absolute path of the awg file to read

    Returns:
        Tuple of instrument settings (a dict), waveforms (list of lists), sequencer settings (list of lists)
    """

    instdict = {}
    waveformlist: List[List[Any]] = [[], []]
    sequencelist: List[List[Any]] = [[], []]
    wfmlen: int

    with open(awgfilepath, 'rb') as fid:

        while True:
            chunk = fid.read(8)
            if not chunk:
                break

            (namelen, valuelen) = struct.unpack('<II', chunk)

            rawname = fid.read(namelen)
            rawvalue = fid.read(valuelen)

            name = rawname[:-1].decode('ascii')  # remove NULL termination char

            if name.startswith('WAVEFORM'):

                namestop = name[name.find('_')+1:].find('_')+name.find('_')
                lookupname = name[:namestop+1]

                if 'DATA' in name:
                    fmtstr = f'{wfmlen}H'
                    AWG_FILE_FORMAT_WAV['WAVEFORM_DATA'] = fmtstr

                file_format = AWG_FILE_FORMAT_WAV[lookupname]
                assert file_format is not None
                value = _unwrap(rawvalue, file_format)
                (number, barename) = _getendingnumber(name)
                fieldname = barename + f"{number-20}"
                waveformlist[0].append(fieldname)
                waveformlist[1].append(value)

                if 'LENGTH' in name:
                    assert isinstance(value, int)
                    wfmlen = value

                continue

            if name.startswith('SEQUENCE'):

                namestop = name[name.find('_')+1:].find('_')+name.find('_')
                lookupname = name[:namestop+1]
                value = _unwrap(rawvalue, AWG_FILE_FORMAT_SEQ[lookupname])
                sequencelist[0].append(name)
                sequencelist[1].append(value)

                continue

            else:
                value = _unwrap(rawvalue, AWG_FILE_FORMAT[name])

            if name in AWG_TRANSLATER:
                assert isinstance(value, int)
                value = AWG_TRANSLATER[name][value]

            instdict.update({name: value})

    return instdict, waveformlist, sequencelist


def _parser2(waveformlist: List[List[Any]]) -> Dict[str, Dict[str, np.ndarray]]:
    """
    Cast the waveformlist from _parser1 into a dict used by _parser3.

    Args:
        waveformlist: A list of lists of waveforms from ``_parser1``

    Returns:
        dict: A dictionary with keys waveform name and values for marker1,
            marker2, and the waveform as np.arrays
    """

    outdict = {}

    for (fieldname, fieldvalue) in zip(waveformlist[0], waveformlist[1]):
        if 'NAME' in fieldname:
            name = fieldvalue
        if 'DATA' in fieldname:
            value = _unpacker(fieldvalue)
            outdict.update({name: {'m1': value[1], 'm2': value[2],
                                   'wfm': value[0]}})

    return outdict


def _parser3(
        sequencelist: List[List[Any]],
        wfmdict: Dict[Any, Any]
             ) -> _parser3_output:
    """
    The final parser! OMG+1
    """

    sequencedict: Dict[str, List[Any]] = {
        'SEQUENCE_WAIT': [],
        'SEQUENCE_LOOP': [],
        'SEQUENCE_JUMP': [],
        'SEQUENCE_GOTO': [],
        'SEQUENCE_WAVEFORM_NAME_CH_1': [],
        'SEQUENCE_WAVEFORM_NAME_CH_2': [],
        'SEQUENCE_WAVEFORM_NAME_CH_3': [],
        'SEQUENCE_WAVEFORM_NAME_CH_4': []
    }

    for fieldname, fieldvalue in zip(sequencelist[0], sequencelist[1]):

        seqnum, name = _getendingnumber(fieldname)

        if 'WAVEFORM' not in name:
            sequencedict[name[:-1]].append(fieldvalue)
        else:
            sequencedict[name[:-1]].append(wfmdict[fieldvalue])

    # clean the dict
    keys = list(sequencedict.keys())
    for key in keys:
        if sequencedict[key] == []:
            sequencedict.pop(key)

    # waveforms
    wfms = []
    m1s = []
    m2s = []
    channels = []
    for key in [key for key in sequencedict if 'WAVE' in key]:
        channels.append(_getendingnumber(key)[0])
        wfms_temp = []
        m1s_temp = []
        m2s_temp = []
        for wfmdict in sequencedict[key]:
            wfms_temp.append(wfmdict['wfm'])
            m1s_temp.append(wfmdict['m1'])
            m2s_temp.append(wfmdict['m2'])

        wfms.append(wfms_temp)
        m1s.append(m1s_temp)
        m2s.append(m2s_temp)

    nreps = sequencedict['SEQUENCE_LOOP']
    waits = sequencedict['SEQUENCE_WAIT']
    gotos = sequencedict['SEQUENCE_GOTO']
    jumps = sequencedict['SEQUENCE_JUMP']

    return wfms, m1s, m2s, nreps, waits, gotos, jumps, channels


def parse_awg_file(
        awgfilepath: str
) -> Tuple[_parser3_output, Dict[str, Union[str, int, Tuple[Any, ...]]]]:
    """
    Parser for a binary .awg file. Returns a tuple matching the call signature
    of make_send_and_load_awg_file and a dictionary with instrument settings

    NOTE: Build-in waveforms are not stored in .awg files. Blame tektronix.

    Args:
        awgfilepath: The absolute path to the awg file

    Returns:
        A tuple and a dict, where the tuple is
        (wfms, m1s, m2s, nreps, trigs, gotos, jumps, channels)
        and the dict contains all instrument settings from the file
    """

    instdict, waveformlist, sequencelist = _parser1(awgfilepath)
    wfmdict = _parser2(waveformlist)
    callsigtuple = _parser3(sequencelist, wfmdict)

    return callsigtuple, instdict



import pytest
import itertools

from qcodes.instrument_drivers.tektronix.Keithley_3706A\
    import Keithley_3706A

import qcodes.instrument.sims as sims
visalib = sims.__file__.replace('__init__.py', 'Keithley_3706A.yaml@sim')


@pytest.fixture(scope='function')
def driver():
    driver = Keithley_3706A('Keithley_3706A',
                            address='GPIB::11::INSTR',
                            visalib=visalib)
    yield driver
    driver.close()


@pytest.fixture()
def channels_in_slot_one():
    channels = []
    for i in range(1, 7):
        for j in range(int(f'1{i}01'), int(f'1{i}17')):
            channels.append(str(j))
    return channels


@pytest.fixture()
def channels_in_slot_two():
    channels = []
    for i in range(1, 7):
        for j in range(int(f'2{i}01'), int(f'2{i}17')):
            channels.append(str(j))
    return channels


@pytest.fixture()
def channels_in_slot_three():
    channels = []
    for i in range(1, 7):
        for j in range(int(f'3{i}01'), int(f'3{i}17')):
            channels.append(str(j))
    return channels


@pytest.fixture()
def matrix_channels():
    slots = ['1', '2', '3']
    rows = [['1', '2', '3', '4', '5', '6']]*3
    columns = [['01', '02', '03', '04', '05', '06', '07', '08', '09',
               '10', '11', '12', '13', '14', '15', '16']]*3
    m_channels = []
    for i, slot in enumerate(slots):
        for element in itertools.product(slot, rows[i], columns[i]):
            m_channels.append(''.join(element))
    return m_channels


def test_idn(driver):
    assert {'firmware': '01.56a',
            'model': '3706A-SNFP',
            'serial': '04447336',
            'vendor': 'KEITHLEY INSTRUMENTS INC.'} == driver.get_idn()


def test_switch_card_idn(driver):
    assert {'slot_no': "1", 'model': "3730",
            'mtype': "6x16 High Density Matrix",
            'firmware': "01.40h",
            'serial': "4447332"} == driver.get_switch_cards()[0]


def test_installed_card_id(driver):
    assert ['1', '2', '3'] == driver._get_slot_ids()


def test_slot_names(driver):
    assert ['slot1', 'slot2', 'slot3'] == driver._get_slot_names()


def test_get_interlock_state(driver):
    dict_list = []
    for i in ['1', '2', '3']:
        dictionary = {'slot_no': i,
                      'state': 'Interlock is disengaged'}
        dict_list.append(dictionary)
    assert tuple(dict_list) == driver.get_interlock_state()


def test_get_number_of_rows(driver):
    assert [6, 6, 6] == driver._get_number_of_rows()


def test_get_number_of_columns(driver):
    assert [16, 16, 16] == driver._get_number_of_columns()


@pytest.mark.parametrize('val', [0, 1, 2])
def test_get_rows(driver, val):
    rows_in_a_slot = ['1', '2', '3', '4', '5', '6']
    assert rows_in_a_slot == driver._get_rows()[val]


@pytest.mark.parametrize('val', [0, 1, 2])
def test_get_columns(driver, val):
    columns_in_a_slot = ['01', '02', '03', '04', '05', '06', '07', '08', '09',
                         '10', '11', '12', '13', '14', '15', '16']
    assert columns_in_a_slot == driver._get_columns()[val]


@pytest.mark.parametrize('val', [1, 2, 3])
def test_number_of_channels_by_slot(driver, val):
    assert 96 == len(driver.get_channels_by_slot(val))


def test_total_number_of_channels(driver):
    assert 3*96 == len(driver.get_channels())


def test_channels_in_slot_one(driver, channels_in_slot_one):
    assert channels_in_slot_one == driver.get_channels_by_slot(1)


def test_channels_in_slot_two(driver, channels_in_slot_two):
    assert channels_in_slot_two == driver.get_channels_by_slot(2)


def test_channels_in_slot_three(driver, channels_in_slot_three):
    assert channels_in_slot_three == driver.get_channels_by_slot(3)


def test_matrix_channels(driver, matrix_channels):
    assert matrix_channels == driver.get_channels()


def test_analog_backplane_specifiers(driver):
    specifiers = ['1911', '1912', '1913', '1914', '1915', '1916',
                  '2911', '2912', '2913', '2914', '2915', '2916',
                  '3911', '3912', '3913', '3914', '3915', '3916']
    assert specifiers == driver.get_analog_backplane_specifiers()


@pytest.mark.parametrize('val', ['slot1', 'allslots', '3111', '3912',
                                 '2103:2116'])
def test_validator_truth(driver, val):
    assert driver._validator(val) is True


@pytest.mark.parametrize('val', ['slot12', 'alslots', '5111', '912', '123213',
                                 'QCoDeS'])
def test_validator_falsehood(driver, val):
    assert driver._validator(val) is False



import pytest

from qcodes.instrument.parameter import Parameter
from qcodes.instrument.sweep_values import SweepValues

from qcodes.utils.validators import Numbers



@pytest.fixture(name='c0')
def _make_c0():
    c0 = Parameter('c0', vals=Numbers(-10, 10), get_cmd=None, set_cmd=None)
    yield c0


@pytest.fixture(name='c1')
def _make_c1():
    c1 = Parameter('c1', get_cmd=None, set_cmd=None)
    yield c1


@pytest.fixture(name='c2')
def _make_c2():
    c2 = Parameter('c2', get_cmd=lambda: 42)
    yield c2


def test_errors(c0, c1, c2):

    # only complete 3-part slices are valid
    with pytest.raises(TypeError):
        c0[1:2]  # For Int params this could be defined as step=1
    with pytest.raises(TypeError):
        c0[:2:3]
    with pytest.raises(TypeError):
        c0[1::3]
    with pytest.raises(TypeError):
        c0[:]  # For Enum params we *could* define this one too...

    # fails if the parameter has no setter
    with pytest.raises(TypeError):
        c2[0:0.1:0.01]

    # validates every step value against the parameter's Validator
    with pytest.raises(ValueError):
        c0[5:15:1]
    with pytest.raises(ValueError):
        c0[5.0:15.0:1.0]
    with pytest.raises(ValueError):
        c0[-12]
    with pytest.raises(ValueError):
        c0[-5, 12, 5]
    with pytest.raises(ValueError):
        c0[-5, 12:8:1, 5]

    # cannot combine SweepValues for different parameters
    with pytest.raises(TypeError):
        c0[0.1] + c1[0.2]

    # improper use of extend
    with pytest.raises(TypeError):
        c0[0.1].extend(5)

    # SweepValue object has no getter, even if the parameter does
    with pytest.raises(AttributeError):
        c0[0.1].get


def test_valid(c0):

    c0_sv = c0[1]
    # setter gets mapped
    assert c0_sv.set == c0.set
    # normal sequence operations access values
    assert list(c0_sv) == [1]
    assert c0_sv[0] == 1
    assert 1 in c0_sv
    assert not (2 in c0_sv)

    # in-place and copying addition
    c0_sv += c0[1.5:1.8:0.1]
    c0_sv2 = c0_sv + c0[2]
    assert list(c0_sv) == [1, 1.5, 1.6, 1.7]
    assert list(c0_sv2) == [1, 1.5, 1.6, 1.7, 2]

    # append and extend
    c0_sv3 = c0[2]
    # append only works with straight values
    c0_sv3.append(2.1)
    # extend can use another SweepValue, (even if it only has one value)
    c0_sv3.extend(c0[2.2])
    # extend can also take a sequence
    c0_sv3.extend([2.3])
    # as can addition
    c0_sv3 += [2.4]
    c0_sv4 = c0_sv3 + [2.5, 2.6]
    assert list(c0_sv3) == [2, 2.1, 2.2, 2.3, 2.4]
    assert list(c0_sv4) == [2, 2.1, 2.2, 2.3, 2.4, 2.5, 2.6]

    # len
    assert len(c0_sv3) == 5

    # in-place and copying reverse
    c0_sv.reverse()
    c0_sv5 = reversed(c0_sv)
    assert list(c0_sv) == [1.7, 1.6, 1.5, 1]
    assert list(c0_sv5) == [1, 1.5, 1.6, 1.7]

    # multi-key init, where first key is itself a list
    c0_sv6 = c0[[1, 3], 4]
    # copying
    c0_sv7 = c0_sv6.copy()
    assert list(c0_sv6) == [1, 3, 4]
    assert list(c0_sv7) == [1, 3, 4]
    assert not (c0_sv6 is c0_sv7)


def test_base():
    p = Parameter('p', get_cmd=None, set_cmd=None)
    with pytest.raises(NotImplementedError):
        iter(SweepValues(p))


def test_snapshot(c0):

    assert c0[0].snapshot() == {
        'parameter': c0.snapshot(),
        'values': [{'item': 0}]
    }

    assert c0[0:5:0.3].snapshot()['values'] == [{
        'first': 0,
        'last': 4.8,
        'num': 17,
        'type': 'linear'
    }]

    sv = c0.sweep(start=2, stop=4, num=5)
    assert sv.snapshot()['values'] == [{
        'first': 2,
        'last': 4,
        'num': 5,
        'type': 'linear'
    }]

    # mixture of bare items, nested lists, and slices
    sv = c0[1, 7, 3.2, [1, 2, 3], 6:9:1, -4.5, 5.3]
    assert sv.snapshot()['values'] == [{
        'first': 1,
        'last': 5.3,
        'min': -4.5,
        'max': 8,
        'num': 11,
        'type': 'sequence'
        }]

    assert (c0[0] + c0[1]).snapshot()['values'] == [
        {'item': 0},
        {'item': 1}
        ]

    assert (c0[0:3:1] + c0[4, 6, 9]).snapshot()['values'] == [
        {'first': 0, 'last': 2, 'num': 3, 'type': 'linear'},
        {'first': 4, 'last': 9, 'min': 4, 'max': 9, 'num': 3,
         'type': 'sequence'}
        ]


def test_repr(c0):
    sv = c0[0]
    assert repr(sv) == (
        f'<qcodes.instrument.sweep_values.SweepFixedValues: c0 at {id(sv)}>'
    )



from typing import Dict, Optional, Any

from .USBHIDMixin import MiniCircuitsHIDMixin
from qcodes.instrument.base import Instrument


class RUDAT_13G_90(Instrument):
    """
    Args:
        name (str)
    """

    def __init__(self, name: str, **kwargs: Any) -> None:

        super().__init__(name, **kwargs)

        self.add_parameter(
            "model_name",
            get_cmd=":MN?"
        )

        self.add_parameter(
            "serial_number",
            get_cmd=":SN?"
        )

        self.add_parameter(
            "firmware",
            get_cmd=":FIRMWARE?"
        )

        self.add_parameter(
            "attenuation",
            set_cmd=":SETATT={}",
            get_cmd=":ATT?",
            get_parser=float
        )

        self.add_parameter(
            "startup_attenuation",
            set_cmd=":STARTUPATT:VALUE:{}",
            get_cmd=":STARTUPATT:VALUE?",
            get_parser=float
        )

        self.add_parameter(
            "hop_points",
            get_cmd="HOP:POINTS?",
            get_parser=int
        )

        self.connect_message()

    def get_idn(self) -> Dict[str, Optional[str]]:
        model = self.model_name()
        serial = self.serial_number()
        firmware = self.firmware()

        return {'vendor': 'Mini-Circuits',
                'model': model,
                'serial': serial,
                'firmware': firmware}


class RUDAT_13G_90_USB(MiniCircuitsHIDMixin, RUDAT_13G_90):
    vendor_id = 0x20ce
    product_id = 0x0023



import logging
import numpy as np
from functools import partial
from typing import Optional, Any, Tuple

from qcodes import VisaInstrument, Instrument
from qcodes import ChannelList, InstrumentChannel
from qcodes.utils import validators as vals
from qcodes.instrument.parameter import (
    MultiParameter,
    ManualParameter,
    ArrayParameter,
    ParamRawDataType,
)
from qcodes.utils.helpers import create_on_off_val_mapping
from qcodes.utils.deprecate import deprecate


log = logging.getLogger(__name__)


class FixedFrequencyTraceIQ(MultiParameter):
    """
    Parameter for sweep that returns the real (I) and imaginary (Q) parts of
    the VNA response.
    Requires the use of the sweep type to be set to continuous wave mode.
    See (https://www.rohde-schwarz.com/webhelp/ZNB_ZNBT_HTML_UserManual_en
    /ZNB_ZNBT_HTML_UserManual_en.htm) under GUI reference -> sweep softtool
    -> sweep type tab -> CW mode
    """

    def __init__(
        self, name: str, instrument: "ZNBChannel", npts: int, bandwidth: int
    ) -> None:
        super().__init__(
            name,
            instrument=instrument,
            names=("I", "Q"),
            labels=(f"{instrument.short_name} I", f"{instrument.short_name} Q"),
            units=("", ""),
            setpoint_names=(
                (f"{instrument.short_name}_frequency",),
                (f"{instrument.short_name}_frequency",),
            ),
            setpoint_units=(("s",), ("s",)),
            setpoint_labels=(("time",), ("time",)),
            shapes=((npts,), (npts,),),
        )
        self.set_cw_sweep(npts, bandwidth)

    def set_cw_sweep(self, npts: int, bandwidth: int) -> None:
        """
        Updates config of the software parameter on sweep change. This is
        needed in order to sync the setpoint shape with the returned data
        shape after a change of sweep settings.

        Sets setpoints to the tuple which are hashable for look up.

        Note: This is similar to the set_sweep functions of the frequency
        sweep parameters. The time setpoints here neglect a small VNA
        overhead. The total time including overhead can be queried with the
        sweep_time function of the vna, but since it is not clear where this
        overhead is spend, we keep the x-axis set to 1/bandwidth. The error
        is only apparent in really fast measurements at 1us and 10us but
        depends on the amount of points you take. More points give less
        overhead.
        """
        t = tuple(np.linspace(0, npts / bandwidth, num=npts))
        self.setpoints = ((t,), (t,))
        self.shapes = ((npts,), (npts,))

    def get_raw(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets the raw real and imaginary part of the data. If parameter
        `cw_check_sweep_first` is set to `True` then at the cost of a few ms
        overhead checks if the vna is setup correctly.
        """
        assert isinstance(self.instrument, ZNBChannel)
        i, q = self.instrument._get_cw_data()
        return i, q


class FixedFrequencyPointIQ(MultiParameter):
    """
    Parameter for sweep that returns the mean of the real (I) and imaginary (Q)
    parts of the VNA response.
    Requires the use of the sweep type to be set to continuous wave mode.
    See (https://www.rohde-schwarz.com/webhelp/ZNB_ZNBT_HTML_UserManual_en
    /ZNB_ZNBT_HTML_UserManual_en.htm) under GUI reference -> sweep softtool
    -> sweep type tab -> CW mode
    Useful for two-tone and other bigger sweeps where you do not want to
    store all individual I-Q values.

    Args:
        name: parameter name
        instrument: instrument the parameter belongs to
    """

    def __init__(self, name: str, instrument: "ZNBChannel") -> None:
        super().__init__(
            name,
            instrument=instrument,
            names=("I", "Q"),
            labels=(f"{instrument.short_name} I", f"{instrument.short_name} Q"),
            units=("", ""),
            setpoints=((), (),),
            shapes=((), (),),
        )

    def get_raw(self) -> Tuple[float, float]:
        """
        Gets the mean of the raw real and imaginary part of the data. If
        parameter `cw_check_sweep_first` is set to `True` then at the cost of a
        few ms overhead checks if the vna is setup correctly.
        """
        assert isinstance(self.instrument, ZNBChannel)
        i, q = self.instrument._get_cw_data()
        return float(np.mean(i)), float(np.mean(q))


class FixedFrequencyPointMagPhase(MultiParameter):
    """
    Parameter for sweep that returns the magnitude of mean of the real (I) and
    imaginary (Q) parts of the VNA response and it's phase.
    Requires the use of the sweep type to be set to continuous wave mode.
    See (https://www.rohde-schwarz.com/webhelp/ZNB_ZNBT_HTML_UserManual_en
    /ZNB_ZNBT_HTML_UserManual_en.htm) under GUI reference -> sweep softtool
    -> sweep type tab -> CW mode

    Args:
        name: parameter name
        instrument: instrument the parameter belongs to
    """

    def __init__(self, name: str, instrument: "ZNBChannel") -> None:
        super().__init__(
            name,
            instrument=instrument,
            names=("magnitude", "phase"),
            labels=(
                f"{instrument.short_name} magnitude",
                f"{instrument.short_name} phase",
            ),
            units=("", "rad"),
            setpoints=((), (),),
            shapes=((), (),),
        )

    def get_raw(self) -> Tuple[float, ...]:
        """
        Gets the magnitude and phase of the mean of the raw real and imaginary
        part of the data. If the parameter `cw_check_sweep_first` is set to
        `True` for the instrument then at the cost of a few ms overhead
        checks if the vna is setup correctly.
        """
        assert isinstance(self.instrument, ZNBChannel)
        i, q = self.instrument._get_cw_data()
        s = np.mean(i) + 1j * np.mean(q)
        return np.abs(s), np.angle(s)


class FrequencySweepMagPhase(MultiParameter):
    """
    Sweep that return magnitude and phase.
    """

    def __init__(
        self,
        name: str,
        instrument: "ZNBChannel",
        start: float,
        stop: float,
        npts: int,
        channel: int,
    ) -> None:
        super().__init__(
            name,
            instrument=instrument,
            names=("magnitude", "phase"),
            labels=(
                f"{instrument.short_name} magnitude",
                f"{instrument.short_name} phase",
            ),
            units=("", "rad"),
            setpoint_units=(("Hz",), ("Hz",)),
            setpoint_labels=(
                (f"{instrument.short_name} frequency",),
                (f"{instrument.short_name} frequency",),
            ),
            setpoint_names=(
                (f"{instrument.short_name}_frequency",),
                (f"{instrument.short_name}_frequency",),
            ),
            shapes=((npts,), (npts,),),
        )
        self.set_sweep(start, stop, npts)
        self._channel = channel

    def set_sweep(self, start: float, stop: float, npts: int) -> None:
        # Needed to update config of the software parameter on sweep change
        # frequency setpoints tuple as needs to be hashable for look up.
        f = tuple(np.linspace(int(start), int(stop), num=npts))
        self.setpoints = ((f,), (f,))
        self.shapes = ((npts,), (npts,))

    def get_raw(self) -> Tuple[ParamRawDataType, ...]:
        assert isinstance(self.instrument, ZNBChannel)
        with self.instrument.format.set_to("Complex"):
            data = self.instrument._get_sweep_data(force_polar=True)
        return abs(data), np.angle(data)


class FrequencySweep(ArrayParameter):
    """
    Hardware controlled parameter class for Rohde Schwarz ZNB trace.

    Instrument returns an array of transmission or reflection data depending
    on the active measurement.

    Args:
        name: parameter name
        instrument: instrument the parameter belongs to
        start: starting frequency of sweep
        stop: ending frequency of sweep
        npts: number of points in frequency sweep

    Methods:
          get(): executes a sweep and returns magnitude and phase arrays

    """

    def __init__(
        self,
        name: str,
        instrument: Instrument,
        start: float,
        stop: float,
        npts: int,
        channel: int,
    ) -> None:
        super().__init__(
            name,
            shape=(npts,),
            instrument=instrument,
            unit="dB",
            label=f"{instrument.short_name} magnitude",
            setpoint_units=("Hz",),
            setpoint_labels=(f"{instrument.short_name} frequency",),
            setpoint_names=(f"{instrument.short_name}_frequency",),
        )
        self.set_sweep(start, stop, npts)
        self._channel = channel

    def set_sweep(self, start: float, stop: float, npts: int) -> None:
        """
        sets the shapes and setpoint arrays of the parameter to
        correspond with the sweep

        Args:
            start: Starting frequency of the sweep
            stop: Stopping frequency of the sweep
            npts: Number of points in the sweep

        """
        # Needed to update config of the software parameter on sweep change
        # freq setpoints tuple as needs to be hashable for look up.
        f = tuple(np.linspace(int(start), int(stop), num=npts))
        self.setpoints = (f,)
        self.shape = (npts,)

    def get_raw(self) -> ParamRawDataType:
        assert isinstance(self.instrument, ZNBChannel)
        return self.instrument._get_sweep_data()


class ZNBChannel(InstrumentChannel):
    def __init__(
        self,
        parent: "ZNB",
        name: str,
        channel: int,
        vna_parameter: Optional[str] = None,
        existing_trace_to_bind_to: Optional[str] = None,
    ) -> None:
        """
        Args:
            parent: Instrument that this channel is bound to.
            name: Name to use for this channel.
            channel: channel on the VNA to use
            vna_parameter: Name of parameter on the vna that this should
                measure such as S12. If left empty this will fall back to
                `name`.
            existing_trace_to_bind_to: Name of an existing trace on the VNA.
                If supplied try to bind to an existing trace with this name
                rather than creating a new trace.

        """
        n = channel
        self._instrument_channel = channel
        # Additional wait when adjusting instrument timeout to sweep time.
        self._additional_wait = 1

        if vna_parameter is None:
            vna_parameter = name
        self._vna_parameter = vna_parameter
        super().__init__(parent, name)

        if existing_trace_to_bind_to is None:
            self._tracename = f"Trc{channel}"
        else:
            traces = self._parent.ask("CONFigure:TRACe:CATalog?")
            if existing_trace_to_bind_to not in traces:
                raise RuntimeError(
                    f"Trying to bind to"
                    f" {existing_trace_to_bind_to} "
                    f"which is not in {traces}"
                )
            self._tracename = existing_trace_to_bind_to

        # map hardware channel to measurement
        # hardware channels are mapped one to one to QCoDeS channels
        # we are not using sub traces within channels.
        if existing_trace_to_bind_to is None:
            self.write(
                f"CALC{self._instrument_channel}:PAR:SDEF"
                f" '{self._tracename}', '{self._vna_parameter}'"
            )

        # Source power is dependent on model, but not well documented.
        # Here we assume -60 dBm for ZNB20, the others are set,
        # due to lack of knowledge, to -80 dBm as of before the edit.
        full_modelname = self._parent.get_idn()["model"]
        if full_modelname is not None:
            model = full_modelname.split("-")[0]
        else:
            raise RuntimeError("Could not determine ZNB model")
        self._model_min_source_power = {
            "ZNB4": -80,
            "ZNB8": -80,
            "ZNB20": -60,
            "ZNB40": -60,
        }
        if model not in self._model_min_source_power.keys():
            raise RuntimeError(f"Unsupported ZNB model: {model}")
        self._min_source_power: float
        self._min_source_power = self._model_min_source_power[model]

        self.add_parameter(
            name="vna_parameter",
            label="VNA parameter",
            get_cmd=f"CALC{self._instrument_channel}:PAR:MEAS? "
                    f"'{self._tracename}'",
            get_parser=self._strip,
        )
        self.add_parameter(
            name="power",
            label="Power",
            unit="dBm",
            get_cmd=f"SOUR{n}:POW?",
            set_cmd=f"SOUR{n}:POW {{:.4f}}",
            get_parser=float,
            vals=vals.Numbers(self._min_source_power, 25),
        )
        self.add_parameter(
            name="bandwidth",
            label="Bandwidth",
            unit="Hz",
            get_cmd=f"SENS{n}:BAND?",
            set_cmd=self._set_bandwidth,
            get_parser=int,
            vals=vals.Enum(
                *np.append(10 ** 6,
                           np.kron([1, 1.5, 2, 3, 5, 7], 10 ** np.arange(6)))
            ),
            docstring="Measurement bandwidth of the IF filter. "
            "The inverse of this sets the integration "
            "time per point. "
            "There is an 'increased bandwidth option' "
            "(p. 4 of manual) that does not get taken "
            "into account here.",
        )
        self.add_parameter(
            name="avg",
            label="Averages",
            unit="",
            get_cmd=f"SENS{n}:AVER:COUN?",
            set_cmd=f"SENS{n}:AVER:COUN {{:.4f}}",
            get_parser=int,
            vals=vals.Ints(1, 5000),
        )
        self.add_parameter(
            name="start",
            get_cmd=f"SENS{n}:FREQ:START?",
            set_cmd=self._set_start,
            get_parser=float,
            vals=vals.Numbers(self._parent._min_freq,
                              self._parent._max_freq - 10),
        )
        self.add_parameter(
            name="stop",
            get_cmd=f"SENS{n}:FREQ:STOP?",
            set_cmd=self._set_stop,
            get_parser=float,
            vals=vals.Numbers(self._parent._min_freq + 1,
                              self._parent._max_freq),
        )
        self.add_parameter(
            name="center",
            get_cmd=f"SENS{n}:FREQ:CENT?",
            set_cmd=self._set_center,
            get_parser=float,
            vals=vals.Numbers(
                self._parent._min_freq + 0.5, self._parent._max_freq - 10
            ),
        )
        self.add_parameter(
            name="span",
            get_cmd=f"SENS{n}:FREQ:SPAN?",
            set_cmd=self._set_span,
            get_parser=float,
            vals=vals.Numbers(1,
                              self._parent._max_freq - self._parent._min_freq),
        )
        self.add_parameter(
            name="npts",
            get_cmd=f"SENS{n}:SWE:POIN?",
            set_cmd=self._set_npts,
            get_parser=int,
        )
        self.add_parameter(
            name="status",
            get_cmd=f"CONF:CHAN{n}:MEAS?",
            set_cmd=f"CONF:CHAN{n}:MEAS {{}}",
            get_parser=int,
        )
        self.add_parameter(
            name="format",
            get_cmd=partial(self._get_format, tracename=self._tracename),
            set_cmd=self._set_format,
            val_mapping={
                "dB": "MLOG\n",
                "Linear Magnitude": "MLIN\n",
                "Phase": "PHAS\n",
                "Unwr Phase": "UPH\n",
                "Polar": "POL\n",
                "Smith": "SMIT\n",
                "Inverse Smith": "ISM\n",
                "SWR": "SWR\n",
                "Real": "REAL\n",
                "Imaginary": "IMAG\n",
                "Delay": "GDEL\n",
                "Complex": "COMP\n",
            },
        )

        self.add_parameter(
            name="trace_mag_phase",
            start=self.start(),
            stop=self.stop(),
            npts=self.npts(),
            channel=n,
            parameter_class=FrequencySweepMagPhase,
        )
        self.add_parameter(
            name="trace",
            start=self.start(),
            stop=self.stop(),
            npts=self.npts(),
            channel=n,
            parameter_class=FrequencySweep,
        )
        self.add_parameter(
            name="electrical_delay",
            label="Electrical delay",
            get_cmd=f"SENS{n}:CORR:EDEL2:TIME?",
            set_cmd=f"SENS{n}:CORR:EDEL2:TIME {{}}",
            get_parser=float,
            unit="s",
        )
        self.add_parameter(
            name="sweep_time",
            label="Sweep time",
            get_cmd=f"SENS{n}:SWE:TIME?",
            get_parser=float,
            unit="s",
        )
        self.add_parameter(
            name="sweep_type",
            get_cmd=f"SENS{n}:SWE:TYPE?",
            set_cmd=self._set_sweep_type,
            val_mapping={
                "Linear": "LIN\n",
                "Logarithmic": "LOG\n",
                "Power": "POW\n",
                "CW_Time": "CW\n",
                "CW_Point": "POIN\n",
                "Segmented": "SEGM\n",
            },
            docstring="The sweep_type parameter is used to set "
            "the type of measurement sweeps. It "
            "allows switching the default linear "
            "VNA sweep type to other types. Note that "
            "at the moment only the linear and "
            "CW_Point modes have supporting "
            "measurement parameters.",
        )
        self.add_parameter(
            name="cw_frequency",
            get_cmd=f"SENS{n}:FREQ:CW?",
            set_cmd=self._set_cw_frequency,
            get_parser=float,
            vals=vals.Numbers(
                self._parent._min_freq + 0.5, self._parent._max_freq - 10
            ),
            docstring="Parameter for setting frequency and "
            "querying for it when VNA sweep type is "
            "set to CW_Point mode.",
        )

        self.add_parameter(
            "cw_check_sweep_first",
            parameter_class=ManualParameter,
            initial_value=True,
            vals=vals.Bool(),
            docstring="Parameter that enables a few commands "
            "which are called before each get in "
            "continuous wave mode checking whether "
            "the vna is setup correctly. Is recommended "
            "to be turned, but can be turned off if "
            "one wants to minimize overhead in fast "
            "measurements. ",
        )

        self.add_parameter(
            name="trace_fixed_frequency",
            npts=self.npts(),
            bandwidth=self.bandwidth(),
            parameter_class=FixedFrequencyTraceIQ,
        )
        self.add_parameter(
            name="point_fixed_frequency", parameter_class=FixedFrequencyPointIQ
        )
        self.add_parameter(
            name="point_fixed_frequency_mag_phase",
            parameter_class=FixedFrequencyPointMagPhase,
        )
        self.add_parameter(
            name="averaging_enabled",
            initial_value=False,
            get_cmd=None,
            set_cmd=self._enable_averaging,
            vals=vals.Bool(),
            val_mapping=create_on_off_val_mapping(on_val="ON", off_val="OFF"),
        )
        self.add_parameter(
            name="auto_sweep_time_enabled",
            initial_value=False,
            get_cmd=None,
            set_cmd=self._enable_auto_sweep_time,
            vals=vals.Bool(),
            val_mapping=create_on_off_val_mapping(on_val="ON", off_val="OFF"),
            docstring="When enabled, the (minimum) sweep time is "
            "calculated internally using the other channel settings "
            "and zero delay",
        )

        self.add_function(
            "set_electrical_delay_auto", call_cmd=f"SENS{n}:CORR:EDEL:AUTO ONCE"
        )
        self.add_function(
            "autoscale",
            call_cmd=f"DISPlay:TRACe1:Y:SCALe:AUTO ONCE, {self._tracename}",
        )

    def _get_format(self, tracename: str) -> str:
        n = self._instrument_channel
        self.write(f"CALC{n}:PAR:SEL '{tracename}'")
        return self.ask(f"CALC{n}:FORM?")

    def _set_format(self, val: str) -> None:
        unit_mapping = {
            "MLOG\n": "dB",
            "MLIN\n": "",
            "PHAS\n": "rad",
            "UPH\n": "rad",
            "POL\n": "",
            "SMIT\n": "",
            "ISM\n": "",
            "SWR\n": "U",
            "REAL\n": "U",
            "IMAG\n": "U",
            "GDEL\n": "S",
            "COMP\n": "",
        }
        label_mapping = {
            "MLOG\n": "Magnitude",
            "MLIN\n": "Magnitude",
            "PHAS\n": "Phase",
            "UPH\n": "Unwrapped phase",
            "POL\n": "Complex Magnitude",
            "SMIT\n": "Complex Magnitude",
            "ISM\n": "Complex Magnitude",
            "SWR\n": "Standing Wave Ratio",
            "REAL\n": "Real Magnitude",
            "IMAG\n": "Imaginary Magnitude",
            "GDEL\n": "Delay",
            "COMP\n": "Complex Magnitude",
        }
        channel = self._instrument_channel
        self.write(f"CALC{channel}:PAR:SEL '{self._tracename}'")
        self.write(f"CALC{channel}:FORM {val}")
        self.trace.unit = unit_mapping[val]
        self.trace.label = f"{self.short_name} {label_mapping[val]}"

    @staticmethod
    def _strip(var: str) -> str:
        """Strip newline and quotes from instrument reply."""
        return var.rstrip()[1:-1]

    def _set_start(self, val: float) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:FREQ:START {val:.7f}")
        stop = self.stop()
        if val >= stop:
            raise ValueError("Stop frequency must be larger than start "
                             "frequency.")
        # we get start as the vna may not be able to set it to the
        # exact value provided.
        start = self.start()
        if abs(val - start) >= 1:
            log.warning(
                "Could not set start to {} setting it to "
                "{}".format(val, start)
            )
        self.update_lin_traces()

    def _set_stop(self, val: float) -> None:
        channel = self._instrument_channel
        start = self.start()
        if val <= start:
            raise ValueError("Stop frequency must be larger than start "
                             "frequency.")
        self.write(f"SENS{channel}:FREQ:STOP {val:.7f}")
        # We get stop as the vna may not be able to set it to the
        # exact value provided.
        stop = self.stop()
        if abs(val - stop) >= 1:
            log.warning(
                "Could not set stop to {} setting it to "
                "{}".format(val, stop)
            )
        self.update_lin_traces()

    def _set_npts(self, val: int) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:SWE:POIN {val:.7f}")
        if self.sweep_type().startswith("CW"):
            self.update_cw_traces()
        else:
            self.update_lin_traces()

    def _set_bandwidth(self, val: int) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:BAND {val:.4f}")
        self.update_cw_traces()

    def _set_span(self, val: float) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:FREQ:SPAN {val:.7f}")
        self.update_lin_traces()

    def _set_center(self, val: float) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:FREQ:CENT {val:.7f}")
        self.update_lin_traces()

    def _set_sweep_type(self, val: str) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:SWE:TYPE {val}")

    def _set_cw_frequency(self, val: float) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:FREQ:CW {val:.7f}")

    def _enable_averaging(self, val: str) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:AVER:STAT {val}")

    def _enable_auto_sweep_time(self, val: str) -> None:
        channel = self._instrument_channel
        self.write(f"SENS{channel}:SWE:TIME:AUTO {val}")

    @deprecate(reason="the method has been renamed",
               alternative="update_lin_traces")
    def update_traces(self) -> None:
        """ updates start, stop and npts of all trace parameters"""
        self.update_lin_traces()

    def update_lin_traces(self) -> None:
        """
        Updates start, stop and npts of all trace parameters
        so that the x-coordinates are updated for the sweep.
        """
        start = self.start()
        stop = self.stop()
        npts = self.npts()
        for _, parameter in self.parameters.items():
            if isinstance(parameter, (FrequencySweep, FrequencySweepMagPhase)):
                try:
                    parameter.set_sweep(start, stop, npts)
                except AttributeError:
                    pass

    def update_cw_traces(self) -> None:
        """
        Updates the bandwidth and npts of all fixed frequency (CW) traces.
        """
        bandwidth = self.bandwidth()
        npts = self.npts()
        for _, parameter in self.parameters.items():
            if isinstance(parameter, FixedFrequencyTraceIQ):
                try:
                    parameter.set_cw_sweep(npts, bandwidth)
                except AttributeError:
                    pass
        self.sweep_time()

    def _get_sweep_data(self, force_polar: bool = False) -> np.ndarray:

        if not self._parent.rf_power():
            log.warning("RF output is off when getting sweep data")
        # It is possible that the instrument and QCoDeS disagree about
        # which parameter is measured on this channel.
        instrument_parameter = self.vna_parameter()
        if instrument_parameter != self._vna_parameter:
            raise RuntimeError(
                "Invalid parameter. Tried to measure "
                f"{self._vna_parameter} "
                f"got {instrument_parameter}"
            )
        self.averaging_enabled(True)
        self.write(f"SENS{self._instrument_channel}:AVER:CLE")

        # preserve original state of the znb
        with self.status.set_to(1):
            self.root_instrument.cont_meas_off()
            try:
                # if force polar is set, the SDAT data format will be used.
                # Here the data will be transferred as a complex number
                # independent of the set format in the instrument.
                if force_polar:
                    data_format_command = "SDAT"
                else:
                    data_format_command = "FDAT"
                timeout = self.sweep_time() + self._additional_wait
                with self.root_instrument.timeout.set_to(timeout):
                    # instrument averages over its last 'avg' number of sweeps
                    # need to ensure averaged result is returned
                    for _ in range(self.avg()):
                        self.write(f"INIT{self._instrument_channel}:IMM; *WAI")
                    self.write(
                        f"CALC{self._instrument_channel}:PAR:SEL "
                        f"'{self._tracename}'"
                    )
                    data_str = self.ask(
                        f"CALC{self._instrument_channel}:DATA?"
                        f" {data_format_command}"
                    )
                data = np.array(data_str.rstrip().split(",")).astype("float64")
                if self.format() in ["Polar",
                                     "Complex",
                                     "Smith",
                                     "Inverse Smith"]:
                    data = data[0::2] + 1j * data[1::2]
            finally:
                self.root_instrument.cont_meas_on()
        return data

    def setup_cw_sweep(self) -> None:
        """
        This method sets the VNA to CW mode. CW Mode sweeps are performed at
        fixed frequency and allow to perform measurements versus time instead
        of versus frequency.
        See (https://www.rohde-schwarz.com/webhelp/ZNB_ZNBT_HTML_UserManual_en
        /ZNB_ZNBT_HTML_UserManual_en.htm) under GUI reference -> sweep softtool
        -> sweep type tab -> CW mode
        """

        # set the channel type to single point msmt
        self.sweep_type("CW_Point")
        # turn off average on the VNA since we want single point sweeps.
        self.averaging_enabled(False)
        # This format is required for getting both real and imaginary parts.
        self.format("Complex")
        # Set the sweep time to auto such that it sets the delay to zero
        # between each point (e.g msmt speed is optimized). Note that if one
        # would like to do a time sweep with time > npts/bandwidth, this is
        # where the delay would be set, but in general we want to measure as
        # fast as possible without artificial delays.
        self.auto_sweep_time_enabled(True)
        # Set cont measurement off here so we don't have to send that command
        # while measuring later.
        self.root_instrument.cont_meas_off()

    def setup_lin_sweep(self) -> None:
        """
        Setup the instrument into linear sweep mode.
        """
        self.sweep_type("Linear")
        self.averaging_enabled(True)
        self.root_instrument.cont_meas_on()

    def _check_cw_sweep(self) -> None:
        """
        Checks if all required settings are met to be able to measure in
        CW_point mode. Similar to what is done in get_sweep_data
        """
        if self.sweep_type() != "CW_Point":
            raise RuntimeError(
                f"Sweep type is not set to continuous wave "
                f"mode, instead it is: {self.sweep_type()}"
            )

        if not self.root_instrument.rf_power():
            log.warning("RF output is off when getting sweep data")

        # It is possible that the instrument and QCoDeS disagree about
        # which parameter is measured on this channel.
        instrument_parameter = self.vna_parameter()
        if instrument_parameter != self._vna_parameter:
            raise RuntimeError(
                "Invalid parameter. Tried to measure "
                f"{self._vna_parameter} "
                f"got {instrument_parameter}"
            )

        # Turn off average on the VNA since we want single point sweeps.
        self.averaging_enabled(False)
        # Set the format to complex.
        self.format("Complex")
        # Set cont measurement off.
        self.root_instrument.cont_meas_off()
        # Cache the sweep time so it is up to date when setting timeouts
        self.sweep_time()

    def _get_cw_data(self) -> Tuple[np.ndarray, np.ndarray]:
        # Make the checking optional such that we can do super fast sweeps as
        # well, skipping the overhead of the other commands.
        if self.cw_check_sweep_first():
            self._check_cw_sweep()

        with self.status.set_to(1):
            timeout = self.sweep_time.cache.get() + self._additional_wait
            with self.root_instrument.timeout.set_to(timeout):
                self.write(f"INIT{self._instrument_channel}:IMM; *WAI")
                data_str = self.ask(f"CALC{self._instrument_channel}:DATA? "
                                    f"SDAT")
            data = np.array(data_str.rstrip().split(",")).astype("float64")
            i = data[0::2]
            q = data[1::2]

        return i, q


class ZNB(VisaInstrument):
    """
    QCoDeS driver for the Rohde & Schwarz ZNB8 and ZNB20
    virtual network analyser. It can probably be extended to ZNB4 and 40
    without too much work.

    Requires FrequencySweep parameter for taking a trace

    Args:
        name: instrument name
        address: Address of instrument probably in format
            'TCPIP0::192.168.15.100::inst0::INSTR'
        init_s_params: Automatically setup channels for all S parameters on the
            VNA.
        reset_channels: If True any channels defined on the VNA at the time
            of initialization are reset and removed.
        **kwargs: passed to base class

    TODO:
    - check initialisation settings and test functions
    """

    CHANNEL_CLASS = ZNBChannel

    def __init__(
        self,
        name: str,
        address: str,
        init_s_params: bool = True,
        reset_channels: bool = True,
        **kwargs: Any,
    ) -> None:

        super().__init__(name=name, address=address, **kwargs)

        # TODO(JHN) I could not find a way to get max and min freq from
        # the API, if that is possible replace below with that
        # See page 1025 in the manual. 7.3.15.10 for details of max/min freq
        # no attempt to support ZNB40, not clear without one how the format
        # is due to variants
        fullmodel = self.get_idn()["model"]
        if fullmodel is not None:
            model = fullmodel.split("-")[0]
        else:
            raise RuntimeError("Could not determine ZNB model")
        # format seems to be ZNB8-4Port
        m_frequency = {
            "ZNB4": (9e3, 4.5e9),
            "ZNB8": (9e3, 8.5e9),
            "ZNB20": (100e3, 20e9),
            "ZNB40": (10e6, 40e9),
        }
        if model not in m_frequency.keys():
            raise RuntimeError(f"Unsupported ZNB model {model}")
        self._min_freq: float
        self._max_freq: float
        self._min_freq, self._max_freq = m_frequency[model]

        self.add_parameter(name="num_ports",
                           get_cmd="INST:PORT:COUN?",
                           get_parser=int)
        num_ports = self.num_ports()

        self.add_parameter(
            name="rf_power",
            get_cmd="OUTP1?",
            set_cmd="OUTP1 {}",
            val_mapping={True: "1\n", False: "0\n"},
        )
        self.add_function("reset", call_cmd="*RST")
        self.add_function("tooltip_on", call_cmd="SYST:ERR:DISP ON")
        self.add_function("tooltip_off", call_cmd="SYST:ERR:DISP OFF")
        self.add_function("cont_meas_on", call_cmd="INIT:CONT:ALL ON")
        self.add_function("cont_meas_off", call_cmd="INIT:CONT:ALL OFF")
        self.add_function("update_display_once", call_cmd="SYST:DISP:UPD ONCE")
        self.add_function("update_display_on", call_cmd="SYST:DISP:UPD ON")
        self.add_function("update_display_off", call_cmd="SYST:DISP:UPD OFF")
        self.add_function(
            "display_sij_split",
            call_cmd=f"DISP:LAY GRID;:DISP:LAY:GRID {num_ports},{num_ports}",
        )
        self.add_function(
            "display_single_window", call_cmd="DISP:LAY GRID;:DISP:LAY:GRID 1,1"
        )
        self.add_function(
            "display_dual_window", call_cmd="DISP:LAY GRID;:DISP:LAY:GRID 2,1"
        )
        self.add_function("rf_off", call_cmd="OUTP1 OFF")
        self.add_function("rf_on", call_cmd="OUTP1 ON")
        if reset_channels:
            self.reset()
            self.clear_channels()
        channels = ChannelList(
            self, "VNAChannels", self.CHANNEL_CLASS, snapshotable=True
        )
        self.add_submodule("channels", channels)
        if init_s_params:
            for i in range(1, num_ports + 1):
                for j in range(1, num_ports + 1):
                    ch_name = "S" + str(i) + str(j)
                    self.add_channel(ch_name)
            self.channels.lock()
            self.display_sij_split()
            self.channels.autoscale()

        self.update_display_on()
        if reset_channels:
            self.rf_off()
        self.connect_message()

    def display_grid(self, rows: int, cols: int) -> None:
        """
        Display a grid of channels rows by columns.
        """
        self.write(f"DISP:LAY GRID;:DISP:LAY:GRID {rows},{cols}")

    def add_channel(self, channel_name: str, **kwargs: Any) -> None:
        i_channel = len(self.channels) + 1
        channel = self.CHANNEL_CLASS(self, channel_name, i_channel, **kwargs)
        self.channels.append(channel)
        if i_channel == 1:
            self.display_single_window()
        if i_channel == 2:
            self.display_dual_window()
        # shortcut
        setattr(self, channel_name, channel)
        # initialising channel
        self.write(f"SENS{i_channel}:SWE:TYPE LIN")
        self.write(f"SENS{i_channel}:SWE:TIME:AUTO ON")
        self.write(f"TRIG{i_channel}:SEQ:SOUR IMM")
        self.write(f"SENS{i_channel}:AVER:STAT ON")

    def clear_channels(self) -> None:
        """
        Remove all channels from the instrument and channel list and
        unlock the channel list.
        """
        self.write("CALCulate:PARameter:DELete:ALL")
        for submodule in self.submodules.values():
            if isinstance(submodule, ChannelList):
                submodule._channels = []
                submodule._channel_mapping = {}
                submodule._locked = False



from qcodes.instrument.delegate.delegate_instrument import (
    GroupedParameter,
    DelegateInstrument
)
from qcodes.instrument.delegate.instrument_group import InstrumentGroup

from qcodes.instrument.delegate.delegate_channel_instrument import (
    DelegateChannelInstrument
)



from functools import partial
from time import time
from typing import Union, cast

from qcodes import ChannelList, InstrumentChannel, VisaInstrument
from qcodes.utils import validators as vals

number = Union[float, int]


class DACException(Exception):
    pass


class DacReader:
    @staticmethod
    def _dac_parse(resp):
        """
        Parses responses from the DAC. They should take the form of
        "<cmd><resp>!" This command returns the value of resp.
        """
        resp = resp.strip()
        if resp[-1] != "!":
            raise DACException(f'Unexpected terminator on response: {resp}. '
                               f'Should end with "!"')
        return resp.strip()[1:-1]

    def _dac_v_to_code(self, volt):
        """
        Convert a voltage to the internal dac code (number between 0-65536)
        based on the minimum/maximum values of a given channel.
        Midrange is 32768.
        """
        if volt < self.min_val or volt >= self.max_val:
            raise ValueError(f'Cannot convert voltage {volt} V ' +
                             'to a voltage code, value out of range '
                             '({} V - {} V).'.format(self.min_val,
                                                     self.max_val))

        frac = (volt - self.min_val) / (self.max_val - self.min_val)
        val = int(round(frac * 65536))
        # extra check to be absolutely sure that the instrument does nothing
        # receive an out-of-bounds value
        if val > 65535 or val < 0:
            raise ValueError('Voltage ({} V) resulted in the voltage code {}'
                             ', which is not within the allowed range.'
                             ''.format(volt, val))
        return val

    def _dac_code_to_v(self, code):
        """
        Convert a voltage to the internal dac code (number between 0-65536)
        based on the minimum/maximum values of a given channel.
        Midrange is 32768.
        """
        frac = code/65536.0
        return (frac * (self.max_val - self.min_val)) + self.min_val

    def _set_slot(self):
        """
        Set the active DAC slot
        """
        resp = self.ask_raw(f"B{self._slot};")
        if int(self._dac_parse(resp)) != self._slot:
            raise DACException("Unexpected return from DAC when setting slot: "
                               f"{resp}. DAC slot may not have been set.")

    def _set_channel(self):
        """
        Set the active DAC channel
        """
        resp = self.ask_raw(f"B{self._slot};C{self._channel};")
        if resp.strip() != f"B{self._slot}!C{self._channel}!":
            raise DACException(f"Unexpected return from DAC when setting "
                               f"channel: {resp}. DAC channel may not have "
                               f"been set.")

    def _query_address(self, addr: int, count: int=1,
                       versa_eeprom: bool=False):
        """
        Query the value at the dac address given.

        Args:
            addr (int): The address to query.

            count (int): The number of bytes to query.

            versa_eeprom(bool): do we want to read from the versadac
            (slot) EEPROM
        """
        # Check if we actually have anything to query
        if count == 0:
            return 0

        # Validate address
        addr = int(addr)
        if addr < 0 or addr > 1107296266:
            raise DACException(f"Invalid address {addr}.")

        # Choose a poke command depending on whether we are querying a
        # VERSADAC eeprom or main memory
        # If we are writing to a VERSADAC, we must also set the slot.
        if versa_eeprom:
            self._set_slot()
            query_command = "e;"
        else:
            query_command = "p;"

        # Read a number of bytes from the device and convert to an int
        val = 0
        for i in range(count):
            # Set DAC to point to address
            ret = int(self._dac_parse(
                self.ask_raw(f"A{addr};")))  # type: ignore[attr-defined]
            if ret != addr:
                raise DACException(f"Failed to set EEPROM address {addr}.")
            val += int(self._dac_parse(self.ask_raw(  # type: ignore[attr-defined]
                query_command))) << (32*(count-i-1))
            addr += 1

        return val

    def _write_address(self, addr: int, val: int,
                       versa_eeprom: bool=False) -> None:
        """
        Write a value to a given DAC address

        Args:
            addr (int): The address to query.

            val (int): The value to write.

            versa_eeprom(bool): do we want to read
             from the versadac (slot) EEPROM
        """
        # Validate address
        addr = int(addr)
        if addr < 0 or addr > 1107296266:
            raise DACException(f"Invalid address {addr}.")

        # Validate value
        val = int(val)
        if val < 0 or val >= 2**32:
            raise DACException(f"Writing invalid value "
                               f"({val}) to address {addr}.")

        # Choose a poke command depending on whether we are querying a
        # VERSADAC eeprom or main memory. If we are writing to a versadac
        # channel we must also set the slot
        if versa_eeprom:
            query_command = "e;"
            write_command = "E"
            self._set_slot()
        else:
            query_command = "p;"
            write_command = "P"

        # Write the value to the DAC
        # Set DAC to point to address
        ret = int(self._dac_parse(self.ask_raw(f"A{addr};")))   # type: ignore[attr-defined]
        if ret != addr:
            raise DACException(f"Failed to set EEPROM address {addr}.")
        self.ask_raw(f"{write_command}{val};")   # type: ignore[attr-defined]
        # Check the write was successful
        if int(self._dac_parse(
                self.ask_raw(query_command))) != val:   # type: ignore[attr-defined]
            raise DACException(f"Failed to write value ({val}) to "
                               f"address {addr}.")


class DacChannel(InstrumentChannel, DacReader):
    """
    A single DAC channel of the DECADAC
    """
    _CHANNEL_VAL = vals.Ints(0, 3)

    def __init__(self, parent, name, channel, min_val=-5, max_val=5):
        super().__init__(parent, name)

        # Validate slot and channel values
        self._CHANNEL_VAL.validate(channel)
        self._channel = channel
        self._slot = self.parent._slot

        # Calculate base address for querying channel parameters
        # Note that the following values can be found using these offsets
        # 0: Interrupt Period
        # 4: DAC High Limit
        # 5: DAC Low Limit
        # 6: Slope (double)
        # 8: DAC Value (double)
        self._base_addr = 1536 + (16*4)*self._slot + 16*self._channel

        # Store min/max voltages
        assert(min_val < max_val)
        self.min_val = min_val
        self.max_val = max_val

        # Add channel parameters
        # Note we will use the older addresses to read the value from the dac
        # rather than the newer 'd' command for backwards compatibility
        self._volt_val = vals.Numbers(self.min_val, self.max_val)
        self.add_parameter(
            "volt",
            get_cmd=partial(self._query_address, self._base_addr + 9, 1),
            get_parser=self._dac_code_to_v,
            set_cmd=self._set_dac,
            set_parser=self._dac_v_to_code,
            vals=self._volt_val,
            label=f"channel {channel+self._slot*4}",
            unit="V",
        )
        # The limit commands are used to sweep dac voltages. They are not
        # safety features.
        self.add_parameter("lower_ramp_limit",
                           get_cmd=partial(self._query_address,
                                           self._base_addr+5),
                           get_parser=self._dac_code_to_v,
                           set_cmd="L{};", set_parser=self._dac_v_to_code,
                           vals=self._volt_val,
                           label="Lower_Ramp_Limit", unit="V")
        self.add_parameter("upper_ramp_limit",
                           get_cmd=partial(self._query_address,
                                           self._base_addr+4),
                           get_parser=self._dac_code_to_v,
                           set_cmd="U{};", set_parser=self._dac_v_to_code,
                           vals=self._volt_val,
                           label="Upper_Ramp_Limit", unit="V")
        self.add_parameter("update_period",
                           get_cmd=partial(self._query_address,
                                           self._base_addr),
                           get_parser=int, set_cmd="T{};", set_parser=int,
                           vals=vals.Ints(50, 65535),
                           label="Update_Period", unit="us")
        self.add_parameter("slope", get_cmd=partial(self._query_address,
                                                    self._base_addr+6, 2),
                           get_parser=int, set_cmd="S{};", set_parser=int,
                           vals=vals.Ints(-(2**32), 2**32),
                           label="Ramp_Slope")

        # Manual parameters to control whether DAC channels should ramp to
        # voltages or jump
        self._ramp_val = vals.Numbers(0, 10)
        self.add_parameter("enable_ramp", get_cmd=None, set_cmd=None,
                           initial_value=False,
                           vals=vals.Bool())
        self.add_parameter("ramp_rate", get_cmd=None, set_cmd=None,
                           initial_value=0.1,
                           vals=self._ramp_val, unit="V/s")

        # Add ramp function to the list of functions
        self.add_function("ramp", call_cmd=self._ramp, args=(self._volt_val,
                                                             self._ramp_val))

        # If we have access to the VERSADAC (slot) EEPROM, we can set the
        # initial value of the channel.
        # NOTE: these values will be overwritten by a K3 calibration
        if self.parent._VERSA_EEPROM_available:
            _INITIAL_ADDR = [6, 8, 32774, 32776]
            self.add_parameter("initial_value",
                               get_cmd=partial(self._query_address,
                                               _INITIAL_ADDR[self._channel],
                                               versa_eeprom=True),
                               get_parser=self._dac_code_to_v,
                               set_cmd=partial(self._write_address,
                                               _INITIAL_ADDR[self._channel],
                                               versa_eeprom=True),
                               set_parser=self._dac_v_to_code,
                               vals=vals.Numbers(self.min_val, self.max_val))

    def _ramp(self, val, rate, block=True):
        """
        Ramp the DAC to a given voltage.

        Params:
            val (float): The voltage to ramp to in volts

            rate (float): The ramp rate in units of volts/s

            block (bool): Should the call block until the ramp is complete?
        """

        # We need to know the current dac value (in raw units), as well as the
        # update rate
        c_volt = self.volt.get()  # Current Voltage
        if c_volt == val:
            # If we are already at the right voltage, we don't need to ramp
            return
        c_val = self._dac_v_to_code(c_volt)  # Current voltage in DAC units
        e_val = self._dac_v_to_code(val)  # Endpoint in DAC units
        # Number of refreshes per second
        t_rate = 1/(self.update_period.get() * 1e-6)
        # Number of seconds to ramp
        secs = abs((c_volt - val)/rate)

        # The formula to calculate the slope is: Number of DAC steps divided by
        # the number of time steps in the ramp multiplied by 65536
        slope = int(((e_val - c_val)/(t_rate*secs))*65536)

        # Now let's set up our limits and ramo slope
        if slope > 0:
            self.upper_ramp_limit.set(val)
        else:
            self.lower_ramp_limit.set(val)
        self.slope.set(slope)

        # Block until the ramp is complete is block is True
        if block:
            while self.slope.get() != 0:
                pass

    def _set_dac(self, code):
        """
        Set the voltage on the dac channel, ramping if the enable_rate
        parameter is set for this channel.

        Params:
            code (int): the DAC code to set the voltage to
        """
        if self.enable_ramp.get():
            self._ramp(self._dac_code_to_v(code), rate=self.ramp_rate.get())
        else:
            code = int(code)
            self._set_channel()
            self.ask_raw(f"U65535;L0;D{code};")

    def write(self, cmd):
        """
        Overload write to set channel prior to any channel operations.
        Since all commands are echoed back, we must keep track of responses
        as well, otherwise commands receive the wrong response.
        """
        self._set_channel()
        return self.ask_raw(cmd)

    def ask(self, cmd):
        """
        Overload ask to set channel prior to operations
        """
        self._set_channel()
        return self.ask_raw(cmd)


class DacSlot(InstrumentChannel, DacReader):
    """
    A single DAC Slot of the DECADAC
    """
    _SLOT_VAL = vals.Ints(0, 4)
    SLOT_MODE_DEFAULT = "Coarse"

    def __init__(self, parent, name, slot, min_val=-5, max_val=5):
        super().__init__(parent, name)

        # Validate slot and channel values
        self._SLOT_VAL.validate(slot)
        self._slot = slot

        # Store whether we have access to the VERSADAC EEPROM
        self._VERSA_EEPROM_available = self.parent._VERSA_EEPROM_available

        # Create a list of channels in the slot
        channels = ChannelList(self, "Slot_Channels", parent.DAC_CHANNEL_CLASS)
        for i in range(4):
            channels.append(parent.DAC_CHANNEL_CLASS(self, f"Chan{i}",
                                                     i, min_val=min_val,
                                                     max_val=max_val))
        self.add_submodule("channels", channels)
        # Set the slot mode. Valid modes are:
        #   Off: Channel outputs are disconnected from the input, grounded
        #       with 10MOhm.
        #   Fine: 2-channel mode. Channels 0 and 1 are output, use 2 and 3
        #       for fine adjustment of Channels 0 and 1 respectively
        #   Coarse: All 4 channels are used as output
        #   FineCald: Calibrated 2-channel mode, with 0 and 1 output, 2 and 3
        #       used automatically for fine adjustment. This mode only works
        #       for calibrated DecaDAC's
        #
        # Unfortunately there is no known way of reading the slot mode hence
        # this will be set in initialization
        if self.parent._cal_supported:
            slot_modes = {"Off": 0, "Fine": 1, "Coarse": 2, "FineCald": 3}
        else:
            slot_modes = {"Off": 0, "Fine": 1, "Coarse": 2}
        self.add_parameter('slot_mode', get_cmd="m;",
                           get_parser=self._dac_parse, set_cmd="M{};",
                           val_mapping=slot_modes)

        # Enable all slots in coarse mode.
        self.slot_mode.set(self.SLOT_MODE_DEFAULT)

    def write(self, cmd):
        """
        Overload write to set channel prior to any channel operations.
        Since all commands are echoed back, we must keep track of responses
        as well, otherwise commands receive the wrong response.
        """
        self._set_slot()
        return self.ask_raw(cmd)

    def ask(self, cmd):
        """
        Overload ask to set channel prior to operations
        """
        self._set_slot()
        return self.ask_raw(cmd)


class Decadac(VisaInstrument, DacReader):
    """
    The qcodes driver for the Decadac.

    Tested with a Decadec firmware revion number 14081 (Decadac 139).

    The message strategy is the following: always keep the queue empty, so
    that self.visa_handle.ask(XXX) will return the answer to XXX and not
    some previous event.


    Attributes:

        _ramp_state (bool): If True, ramp state is ON. Default False.

        _ramp_time (int): The ramp time in ms. Default 100 ms.
    """
    DAC_CHANNEL_CLASS = DacChannel
    DAC_SLOT_CLASS = DacSlot

    def __init__(self, name: str, address: str,
                 min_val: number=-5, max_val: number=5,
                 **kwargs) -> None:
        """

        Creates an instance of the Decadac instruments

        Args:
            name: What this instrument is called locally.

            address: The address of the DAC. For a serial port this
                is ASRLn::INSTR where n is replaced with the address set in the
                VISA control panel. Baud rate and other serial parameters must
                also be set in the VISA control panel.

            min_val: The minimum value in volts that can be output by the DAC.
                This value should correspond to the DAC code 0.

            max_val: The maximum value in volts that can be output by the DAC.
                This value should correspond to the DAC code 65536.

        """

        super().__init__(name, address, **kwargs)

        # Do feature detection
        self._feature_detect()

        # Create channels
        channels = ChannelList(self, "Channels", self.DAC_CHANNEL_CLASS,
                               snapshotable=False)
        slots = ChannelList(self, "Slots", self.DAC_SLOT_CLASS)
        for i in range(5):  # Create the 6 DAC slots
            slots.append(self.DAC_SLOT_CLASS(self, f"Slot{i}", i,
                                             min_val, max_val))
            slot_channels = slots[i].channels
            slot_channels = cast(ChannelList, slot_channels)
            channels.extend(slot_channels)
        slots.lock()
        channels.lock()
        self.add_submodule("slots", slots)
        self.add_submodule("channels", channels)

        self.connect_message()

    def set_all(self, volt: float) -> None:
        """
        Set all dac channels to a specific voltage. If channels are set to ramp
        then the ramps will occur in sequence, not simultaneously.

        Args:
            volt(float): The voltage to set all gates to.
        """
        for chan in self.channels:
            chan.volt.set(volt)

    def ramp_all(self, volt, ramp_rate):
        """
        Ramp all dac channels to a specific voltage at the given rate
        simultaneously. Note that the ramps are not synchronized due to
        communications time and DAC ramps starting as soon as the commands are
        in.

        Args:
            volt(float): The voltage to ramp all channels to.

            ramp_rate(float): The rate in volts per second to ramp
        """
        # Start all channels ramping
        for chan in self.channels:
            chan._ramp(volt, ramp_rate, block=False)

        # Wait for all channels to complete ramping.
        # The slope is reset to 0 once ramping is complete.
        for chan in self.channels:
            while chan.slope.get():
                pass

    def get_idn(self):
        """
        Attempt to identify the dac. Since we don't have standard SCPI
        commands, ``*IDN`` will do nothing on this DAC.

        Returns:
            A dict containing a serial and hardware version
        """
        self._feature_detect()

        return {"serial": self.serial_no, "hardware_version": self.version}

    def connect_message(self, idn_param='IDN', begin_time=None):
        """
        Print a connect message, taking into account the lack of a standard
        ``*IDN`` on the Harvard DAC

        Args:
            begin_time (int, float): time.time() when init started.
                Default is self._t0, set at start of Instrument.__init__.
        """
        # start with an empty dict, just in case an instrument doesn't
        # heed our request to return all 4 fields.
        t = time() - (begin_time or self._t0)

        con_msg = ("Connected to Harvard DecaDAC "
                   "(hw ver: {}, serial: {}) in {:.2f}s".format(
                    self.version, self.serial_no, t))
        print(con_msg)

    def __repr__(self):
        """Simplified repr giving just the class and name."""
        return f"<{type(self).__name__}: {self.name}>"

    def _feature_detect(self):
        """
        Detect which features are available on the DAC by querying various
        parameters.
        """

        # Check whether EEPROM is installed
        try:
            if self._query_address(1107296256) == 21930:
                self._EEPROM_available = True
            else:
                self._EEPROM_available = False
        except DACException:
            self._EEPROM_available = False

        # Check whether we can set startup values for the DAC.
        # This requires access to the EEPROM on each slot

        # note from DV: the value never gets set to True in this driver.
        # To avoid an error of a non existing attribute, here I set it to
        # False by default
        self._VERSA_EEPROM_available = False

        try:
            # Let's temporarily pretend to be slot 0
            self._slot = 0
            self._query_address(6, versa_eeprom=True)
            del self._slot
        except DACException:
            pass

        # Check whether calibration is supported
        try:
            if self._dac_parse(self.ask("k;")):
                self._cal_supported = True
        except DACException:
            self._cal_supported = False

        # Finally try and read the DAC version and S/N.
        # This is only possible if the EEPROM is queryable.
        if self._EEPROM_available:
            self.version = self._query_address(1107296266)
            self.serial_no = self._query_address(1107296264)
        else:
            self.version = 0
            self.serial_no = 0

    def write(self, cmd):
        """
        Since all commands are echoed back, we must keep track of responses
        as well, otherwise commands receive the wrong response. Hence
        all writes must also read a response.
        """
        return self.ask(cmd)



"""
IO managers for QCodes.

IO managers wrap whatever physical storage layer the user wants to use
in an interface mimicking the built-in <open> context manager, with
some restrictions to minimize the overhead in creating new IO managers.

The main thing these managers need to implement is the open context manager:

- Only the context manager needs to be implemented, not separate
  open function and close methods.

- open takes the standard parameters:

    - filename: (string)
    - mode: (string) only 'r' (read), 'w' (write), and 'a' (append) are
      expected to be implemented. As with normal file objects, the only
      difference between write and append is that write empties the file
      before adding new data, and append leaves the existing contents in
      place but starts writing at the end.
    - encoding: If a special output encoding is desired. i.e. 'utf8

- the file-like object returned should implement a minimal set of operations.

  In read mode:
    - read([size]): read to the end or at most size bytes into a string
    - readline([size]): read until a newline or up to size bytes, into a string
    - iter(): usually return self, but can be any iterator over lines
    - next(): assuming iter() returns self, this yields the next line.

  In write or append mode:
    - write(s): add string s to the end of the file.
    - writelines(seq): add a sequence of strings

IO managers should also implement:

- a join method, ala ``os.path.join(*args)``.
- a list method, that returns all objects matching location
- a remove method, ala os.remove(path) except that it will remove directories
  as well as files, since we're allowing "locations" to be directories
  or files.
"""

import os
import re
import shutil
from contextlib import contextmanager
from fnmatch import fnmatch

ALLOWED_OPEN_MODES = ('r', 'w', 'a')


class DiskIO:

    """
    Simple IO object to wrap disk operations with a custom base location.

    Also accepts both forward and backward slashes at any point, and
    normalizes both to the OS we are currently on.

    Args:
        base_location (str): a path to the root data folder.
            Converted to an absolute path immediately, so even if you supply a
            relative path, later changes to the OS working directory will not
            affect data paths.
    """

    def __init__(self, base_location):
        if base_location is None:
            self.base_location = None
        else:
            base_location = self._normalize_slashes(base_location)
            self.base_location = os.path.abspath(base_location)

    @contextmanager
    def open(self, filename, mode, encoding=None):
        """
        Mimic the interface of the built in open context manager.

        Args:
            filename (str): path relative to base_location.

            mode (str): 'r' (read), 'w' (write), or 'a' (append).
                Other open modes are not supported because we don't want
                to force all IO managers to support others.

        Returns:
            context manager yielding the open file
        """
        if mode not in ALLOWED_OPEN_MODES:
            raise ValueError(f'mode {mode} not allowed in IO managers')

        filepath = self.to_path(filename)

        # make directories if needed
        dirpath = os.path.dirname(filepath)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        # normally we'd construct this context manager with try/finally, but
        # here we already have a context manager for open so we just wrap it
        with open(filepath, mode, encoding=encoding) as f:
            yield f

    def _normalize_slashes(self, location):
        # note that this is NOT os.path.join - the difference is os.path.join
        # discards empty strings, so if you use it on a re.split absolute
        # path you will get a relative path!
        return os.sep.join(re.split('[\\\\/]', location))

    def to_path(self, location):
        """
        Convert a location string into a path on the local file system.

        For DiskIO this just fixes slashes and prepends the base location,
        doing nothing active with the file. But for other io managers that
        refer to remote storage, this method may actually fetch the file and
        put it at a temporary local path.

        Args:
            location (str): A location string for a complete dataset or
                a file within it.

        Returns:
            str: The path on disk to which this location maps.
        """
        location = self._normalize_slashes(location)
        if self.base_location:
            return os.path.join(self.base_location, location)
        else:
            return location

    def to_location(self, path):
        """
        Convert a local filesystem path into a location string.

        Args:
            path (str): a path on the local file system.

        Returns:
            str: the location string corresponding to this path.
        """
        if self.base_location:
            return os.path.join(self.base_location, path)
        else:
            return path

    def __repr__(self):
        """Show the base location in the repr."""
        return f"<DiskIO, base_location={repr(self.base_location)}>"

    def join(self, *args):
        """Context-dependent os.path.join for this io manager."""
        return os.path.join(*list(map(self._normalize_slashes, args)))

    def isfile(self, location):
        """Check whether this location matches a file."""
        path = self.to_path(location)
        return os.path.isfile(path)

    def list(self, location, maxdepth=1, include_dirs=False):
        """
        Return all files that match location.

        This is either files whose names match up to an arbitrary extension,
        or any files within an exactly matching directory name.

        Args:
            location (str): the location to match.
                May contain the usual path wildcards * and ?

            maxdepth (Optional[int]): maximum levels of directory nesting to
                recurse into looking for files. Default 1.

            include_dirs (Optional[bool]): whether to allow directories in
                the results or just files. Default False.

        Returns:
            A list of matching files and/or directories, as locations
            relative to our base_location.
        """
        location = self._normalize_slashes(location)
        search_dir, pattern = os.path.split(location)
        path = self.to_path(search_dir)

        if not os.path.isdir(path):
            return []

        matches = [fn for fn in os.listdir(path) if fnmatch(fn, pattern + '*')]
        out = []

        for match in matches:
            matchpath = self.join(path, match)
            if os.path.isdir(matchpath) and fnmatch(match, pattern):
                if maxdepth > 0:
                    # exact directory match - walk down to maxdepth
                    for root, dirs, files in os.walk(matchpath, topdown=True):
                        depth = root[len(path):].count(os.path.sep)
                        if depth == maxdepth:
                            dirs[:] = []  # don't recurse any further

                        for fn in files + (dirs if include_dirs else []):
                            out.append(self.to_location(self.join(root, fn)))

                elif include_dirs:
                    out.append(self.join(search_dir, match))

            elif (os.path.isfile(matchpath) and
                  (fnmatch(match, pattern) or
                   fnmatch(os.path.splitext(match)[0], pattern))):
                # exact filename match, or match up to an extension
                # note that we need fnmatch(match, pattern) in addition to the
                # splitext test to cover the case of the base filename itself
                # containing a dot.
                out.append(self.join(search_dir, match))

        return out

    def remove(self, filename):
        """Delete a file or folder and prune the directory tree."""
        path = self.to_path(filename)
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)

        filepath = os.path.split(path)[0]
        try:
            os.removedirs(filepath)
        except OSError:
            # directory was not empty - good that we're not removing it!
            pass

    def remove_all(self, location):
        """
        Delete all files/directories in the dataset at this location.

        Afterward prunes the directory tree.
        """
        for fn in self.list(location):
            self.remove(fn)



import unittest
from typing import Optional
"""
This module defines:

- `DriverTestCase`: a `TestCase` subclass meant for testing instrument drivers

- `test_instrument`: a function to test one instrument, given its test class

- `test_instruments`: a function to test all instruments that have been defined
                      in your python session


Using `DriverTestCase` is pretty easy:

- Inherit from this class instead of from the base `unittest.TestCase`

- Provide a driver class variable that points to the Instrument class

- In your tests, `self.instrument` is the latest instance of this class.

- If your test case includes a `setUpClass` method, make sure to call
  `super().setUpClass()`, because that's where we find the latest instance of
  this `Instrument`, or skip the test case if no instances are found.
"""


class DriverTestCase(unittest.TestCase):
    # override this in a subclass
    driver = None  # type: Optional[type]

    @classmethod
    def setUpClass(cls):
        if cls is DriverTestCase:
            return

        if cls.driver is None:
            raise TypeError('you must set a driver for ' + cls.__name__)

        instances = cls.driver.instances()
        name = cls.driver.__name__

        if not instances:
            msg = f'no instances of {name} found'
            if getattr(cls, 'noskip', False):
                # just to test this class, we need to disallow skipping
                raise ValueError(msg)
            else:
                raise unittest.SkipTest(msg)

        if len(instances) == 1:
            print(f'***** found one {name}, testing *****')
        else:
            print('***** found {} instances of {}; '
                  'testing the last one *****'.format(len(instances), name))

        cls.instrument = instances[-1]


def test_instruments(verbosity=1):
    """
    Discover available instruments and test them all
    Unlike test_instrument, this does NOT reload tests prior to running them

    optional verbosity (default 1)
    """
    import qcodes.instrument_drivers as qcdrivers
    import qcodes

    driver_path = qcdrivers.__path__[0]
    suite = unittest.defaultTestLoader.discover(
        driver_path, top_level_dir=qcodes.__path__[0])
    unittest.TextTestRunner(verbosity=verbosity).run(suite)


def test_instrument(instrument_testcase, verbosity=2):
    """
    Runs one instrument testcase
    Reloads the test case before running it

    optional verbosity (default 2)
    """
    import sys
    import importlib

    # reload the test case
    module_name = instrument_testcase.__module__
    class_name = instrument_testcase.__name__
    del sys.modules[module_name]

    module = importlib.import_module(module_name)
    reloaded_testcase = getattr(module, class_name)

    suite = unittest.defaultTestLoader.loadTestsFromTestCase(reloaded_testcase)
    unittest.TextTestRunner(verbosity=verbosity).run(suite)



"""
This module holds the objects that describe the intra-run relationships
between the parameters of that run. Most importantly, the information about
which parameters depend on each other is handled here.
"""
from copy import deepcopy
from typing import (Any, Dict, FrozenSet, Iterable, List, Optional, Sequence,
                    Set, Tuple, Type)

from .param_spec import ParamSpec, ParamSpecBase
from .versioning.rundescribertypes import InterDependencies_Dict

ParamSpecTree = Dict[ParamSpecBase, Tuple[ParamSpecBase, ...]]
ParamNameTree = Dict[str, Tuple[str, ...]]
ErrorTuple = Tuple[Type[Exception], str]


class DependencyError(Exception):
    def __init__(self,
                 param_name: str,
                 missing_params: Set[str],
                 *args: Any):
        super().__init__(*args)
        self._param_name = param_name
        self._missing_params = missing_params

    def __str__(self) -> str:
        return (f'{self._param_name} has the following dependencies that are '
                f'missing: {self._missing_params}')


class InferenceError(Exception):
    def __init__(self,
                 param_name: str,
                 missing_params: Set[str],
                 *args: Any):
        super().__init__(*args)
        self._param_name = param_name
        self._missing_params = missing_params

    def __str__(self) -> str:
        return (f'{self._param_name} has the following inferences that are '
                f'missing: {self._missing_params}')


class InterDependencies_:
    """
    Object containing a group of ParamSpecs and the information about their
    internal relations to each other
    """

    def __init__(self,
                 dependencies: Optional[ParamSpecTree] = None,
                 inferences: Optional[ParamSpecTree] = None,
                 standalones: Tuple[ParamSpecBase, ...] = ()):

        dependencies = dependencies or {}
        inferences = inferences or {}

        deps_error = self.validate_paramspectree(dependencies)
        if deps_error is not None:
            old_error = deps_error[0](deps_error[1])
            raise ValueError('Invalid dependencies') from old_error

        inffs_error = self.validate_paramspectree(inferences)
        if inffs_error is not None:
            old_error = inffs_error[0](inffs_error[1])
            raise ValueError('Invalid inferences') from old_error

        link_error = self._validate_double_links(dependencies, inferences)
        if link_error is not None:
            error, mssg = link_error
            raise error(mssg)

        for ps in standalones:
            if not isinstance(ps, ParamSpecBase):
                base_error = TypeError('Standalones must be a sequence of '
                                       'ParamSpecs')

                raise ValueError('Invalid standalones') from base_error

        self._remove_duplicates(dependencies)
        self._remove_duplicates(inferences)

        self.dependencies: ParamSpecTree = dependencies
        self.inferences: ParamSpecTree = inferences
        self.standalones: FrozenSet[ParamSpecBase] = frozenset(standalones)

        # The object is now complete, but for convenience, we form some more
        # private attributes for easy look-up

        self._id_to_paramspec: Dict[str, ParamSpecBase] = {}
        for tree in (self.dependencies, self.inferences):
            for ps, ps_tup in tree.items():
                self._id_to_paramspec.update({ps.name: ps})
                self._id_to_paramspec.update({pst.name: pst
                                              for pst in ps_tup})
        for ps in self.standalones:
            self._id_to_paramspec.update({ps.name: ps})
        self._paramspec_to_id = {v: k for k, v in self._id_to_paramspec.items()}

        self._dependencies_inv: ParamSpecTree = self._invert_tree(
            self.dependencies)
        self._inferences_inv: ParamSpecTree = self._invert_tree(
            self.inferences)

        # Set operations are ~2x (or more) faster on strings than on hashable
        # ParamSpecBase objects, hence the need for use of the following
        # representation
        self._deps_names: ParamNameTree = self._tree_of_names(
            self.dependencies)
        self._infs_names: ParamNameTree = self._tree_of_names(
            self.inferences)

    @staticmethod
    def _tree_of_names(tree: ParamSpecTree) -> ParamNameTree:
        """
        Helper function to convert a ParamSpecTree-kind of tree where all
        ParamSpecBases are substituted with their ``.name`` s.
        Will turn {A: (B, C)} into {A.name: (B.name, C.name)}
        """
        name_tree: ParamNameTree = {}
        for ps, ps_tuple in tree.items():
            name_tree[ps.name] = tuple(p.name for p in ps_tuple)
        return name_tree

    @staticmethod
    def _invert_tree(tree: ParamSpecTree) -> ParamSpecTree:
        """
        Helper function to invert a ParamSpecTree. Will turn {A: (B, C)} into
        {B: (A,), C: (A,)}
        """
        indeps: Set[ParamSpecBase] = set()
        for indep_tup in tree.values():
            indeps.update(indep_tup)

        inverted: ParamSpecTree = {}
        for indep in indeps:
            deps = tuple(ps for ps in tree if indep in tree[ps])
            inverted[indep] = deps

        return inverted

    @staticmethod
    def _remove_duplicates(tree: ParamSpecTree) -> None:
        """
        Helper function to remove duplicate entries from a ParamSpecTree while
        preserving order. Will turn {A: (B, B, C)} into {A: (B, C)}
        """
        for ps, tup in tree.items():
            specs: List[ParamSpecBase] = []
            for p in tup:
                if p not in specs:
                    specs.append(p)
            tree[ps] = tuple(specs)

    @staticmethod
    def validate_paramspectree(
            paramspectree: ParamSpecTree) -> Optional[ErrorTuple]:
        """
        Validate a ParamSpecTree. Apart from adhering to the type, a
        ParamSpecTree must not have any cycles.

        Returns:
            A tuple with an exception type and an error message or None, if
            the paramtree is valid
        """

        # Validate the type

        if not isinstance(paramspectree, dict):
            return (TypeError, 'ParamSpecTree must be a dict')

        for key, values in paramspectree.items():
            if not isinstance(key, ParamSpecBase):
                return (TypeError, 'ParamSpecTree must have ParamSpecs as keys')
            if not isinstance(values, tuple):
                return (TypeError, 'ParamSpecTree must have tuple values')
            for value in values:
                if not isinstance(value, ParamSpecBase):
                    return (TypeError,
                            ('ParamSpecTree can only have tuples of '
                             'ParamSpecs as values'))

        # check for cycles

        roots = set(paramspectree.keys())
        leafs = {ps for tup in paramspectree.values() for ps in tup}

        if roots.intersection(leafs) != set():
            return (ValueError, 'ParamSpecTree can not have cycles')

        return None

    @staticmethod
    def _validate_double_links(tree1: ParamSpecTree,
                               tree2: ParamSpecTree) -> Optional[ErrorTuple]:
        """
        Validate that two trees do not contain double links. A double link
        is a link between two nodes that exists in both trees. E.g. if the
        first tree has {A: (B, C)}, the second may not have {B: (A,)} etc.
        """
        for ps, tup in tree1.items():
            for val in tup:
                if ps in tree2.get(val, ()):
                    mssg = (f"Invalid dependencies/inferences. {ps} and "
                            f"{val} have an ill-defined relationship.")
                    return (ValueError, mssg)

        return None

    def what_depends_on(self, ps: ParamSpecBase) -> Tuple[ParamSpecBase, ...]:
        """
        Return a tuple of the parameters that depend on the given parameter.
        Returns an empty tuple if nothing depends on the given parameter

        Args:
            ps: the parameter to look up

        Raises:
            ValueError if the parameter is not part of this object
        """
        if ps not in self:
            raise ValueError(f'Unknown parameter: {ps}')
        return self._dependencies_inv.get(ps, ())

    def what_is_inferred_from(self,
                              ps: ParamSpecBase) -> Tuple[ParamSpecBase, ...]:
        """
        Return a tuple of the parameters thatare inferred from the given
        parameter. Returns an empty tuple if nothing is inferred from the given
        parameter

        Args:
            ps: the parameter to look up

        Raises:
            ValueError if the parameter is not part of this object
        """
        if ps not in self:
            raise ValueError(f'Unknown parameter: {ps}')
        return self._inferences_inv.get(ps, ())

    def _to_dict(self) -> InterDependencies_Dict:
        """
        Write out this object as a dictionary
        """
        parameters = {key: value._to_dict() for key, value in
                      self._id_to_paramspec.items()}
        dependencies = self._construct_subdict('dependencies')
        inferences = self._construct_subdict('inferences')
        standalones = [self._paramspec_to_id[ps] for ps in
                       self.standalones]
        output: InterDependencies_Dict = {"parameters": parameters,
                                          "dependencies": dependencies,
                                          "inferences": inferences,
                                          "standalones": standalones}
        return output

    def _construct_subdict(self, treename: str) -> Dict[str, Any]:
        output = {}
        for key, value in getattr(self, treename).items():
            ps_id = self._paramspec_to_id[key]
            ps_ids = [self._paramspec_to_id[ps] for ps in value]
            output.update({ps_id: ps_ids})
        return output

    @property
    def paramspecs(self) -> Tuple[ParamSpecBase, ...]:
        """
        Return the ParamSpecBase objects of this instance
        """
        return tuple(self._paramspec_to_id.keys())

    @property
    def non_dependencies(self) -> Tuple[ParamSpecBase, ...]:
        """
        Return all parameters that are not dependencies of other parameters,
        i.e. return the top level parameters. Returned tuple is sorted by
        parameter names.
        """
        non_dependencies = tuple(self.standalones) \
                         + tuple(self.dependencies.keys())
        non_dependencies_sorted_by_name = tuple(
            sorted(non_dependencies, key=lambda ps: ps.name))
        return non_dependencies_sorted_by_name

    @property
    def names(self) -> Tuple[str, ...]:
        """
        Return all the names of the parameters of this instance
        """
        return tuple(self._id_to_paramspec.keys())

    def _extend_with_paramspec(self, ps: ParamSpec) -> 'InterDependencies_':
        """
        Create a new InterDependencies_ object extended with the provided
        ParamSpec. A helper function for DataSet's add_parameter function.
        Note that this function will only work as expected if the ParamSpecs
        are extended into the InterDependencies_ in the "logical order", i.e.
        independent ParamSpecs before dependent ones.
        """
        base_ps = ps.base_version()

        old_standalones = set(self.standalones.copy())
        new_standalones: Tuple[ParamSpecBase, ...]

        if len(ps.depends_on_) > 0:
            deps_list = [self._id_to_paramspec[name] for name in ps.depends_on_]
            new_deps = {base_ps: tuple(deps_list)}
            old_standalones = old_standalones.difference(set(deps_list))
        else:
            new_deps = {}

        if len(ps.inferred_from_) > 0:
            inffs_list = [self._id_to_paramspec[name]
                          for name in ps.inferred_from_]
            new_inffs = {base_ps: tuple(inffs_list)}
            old_standalones = old_standalones.difference(set(inffs_list))
        else:
            new_inffs = {}

        if new_deps == new_inffs == {}:
            new_standalones = (base_ps,)
        else:
            old_standalones = old_standalones.difference({base_ps})
            new_standalones = ()

        new_deps.update(self.dependencies.copy())
        new_inffs.update(self.inferences.copy())
        new_standalones = tuple(list(new_standalones) + list(old_standalones))

        new_idps = InterDependencies_(dependencies=new_deps,
                                      inferences=new_inffs,
                                      standalones=new_standalones)

        return new_idps

    def extend(
            self,
            dependencies: Optional[ParamSpecTree] = None,
            inferences: Optional[ParamSpecTree] = None,
            standalones: Tuple[ParamSpecBase, ...] = ()) -> 'InterDependencies_':
        """
        Create a new InterDependencies_ object that is an extension of this
        instance with the provided input
        """

        dependencies = {} if dependencies is None else dependencies
        inferences = {} if inferences is None else inferences

        # first step: remove parameters from standalones if they no longer
        # stand alone

        depended_on = (ps for tup in dependencies.values() for ps in tup)
        inferred_from = (ps for tup in inferences.values() for ps in tup)

        standalones_mut = set(deepcopy(self.standalones))
        standalones_mut = (standalones_mut.difference(set(dependencies))
                                          .difference(set(inferences))
                                          .difference(set(depended_on))
                                          .difference(set(inferred_from)))

        # then update deps and inffs
        new_deps = deepcopy(self.dependencies)
        for ps in set(dependencies).intersection(set(new_deps)):
            new_deps[ps] = tuple(set(list(new_deps[ps]) +
                                     list(dependencies[ps])))
        for ps in set(dependencies).difference(set(new_deps)):
            new_deps.update({deepcopy(ps): dependencies[ps]})

        new_inffs = deepcopy(self.inferences.copy())
        for ps in set(inferences).intersection(set(new_inffs)):
            new_inffs[ps] = tuple(set(list(new_inffs[ps]) +
                                      list(inferences[ps])))
        for ps in set(inferences).difference(set(new_inffs)):
            new_inffs.update({deepcopy(ps): inferences[ps]})

        # add new standalones
        new_standalones = tuple(standalones_mut.union(set(standalones)))

        new_idps = InterDependencies_(dependencies=new_deps,
                                      inferences=new_inffs,
                                      standalones=new_standalones)

        return new_idps

    def remove(self, parameter: ParamSpecBase) -> 'InterDependencies_':
        """
        Create a new InterDependencies_ object that is similar to this
        instance, but has the given parameter removed.
        """
        if parameter not in self:
            raise ValueError(f'Unknown parameter: {parameter}.')

        if parameter in self._dependencies_inv:
            raise ValueError(f'Cannot remove {parameter.name}, other '
                             'parameters depend on it.')
        if parameter in self._inferences_inv:
            raise ValueError(f'Cannot remove {parameter.name}, other '
                             'parameters are inferred from it.')

        if parameter in self.standalones:
            new_standalones = tuple(deepcopy(self.standalones).
                                    difference({parameter}))
            new_deps = deepcopy(self.dependencies)
            new_inffs = deepcopy(self.inferences)
        elif parameter in self.dependencies or parameter in self.inferences:
            new_deps = deepcopy(self.dependencies)
            new_inffs = deepcopy(self.inferences)
            # figure out whether removing this parameter will make any other
            # parameters standalone
            new_standalones_l = []
            old_standalones = deepcopy(self.standalones)
            for indep in self.dependencies.get(parameter, []):
                if not(indep in self._inferences_inv or
                       indep in self.inferences):
                    new_standalones_l.append(indep)
            for basis in self.inferences.get(parameter, []):
                if not(basis in self._dependencies_inv or
                       basis in self.dependencies):
                    new_standalones_l.append(basis)
            new_deps.pop(parameter, None)
            new_inffs.pop(parameter, None)
            new_standalones = tuple(set(new_standalones_l)
                                    .union(old_standalones))
        else:
            raise ValueError(f'Inconsistent InterDependencies_ object '
                             f'parameter: {parameter} is in list of'
                             f'parameters but is neither a "standalone", '
                             f'"dependency" or "inference"')

        idps = InterDependencies_(dependencies=new_deps, inferences=new_inffs,
                                  standalones=new_standalones)
        return idps

    def validate_subset(self, parameters: Sequence[ParamSpecBase]) -> None:
        """
        Validate that the given parameters form a valid subset of the
        parameters of this instance, meaning that all the given parameters are
        actually found in this instance and that there are no missing
        dependencies/inferences.

        Args:
            parameters: The collection of ParamSpecBases to validate

        Raises:
            DependencyError, if a dependency is missing
            InferenceError, if an inference is missing
        """
        params = {p.name for p in parameters}

        for param in params:
            ps = self._id_to_paramspec.get(param, None)
            if ps is None:
                raise ValueError(f'Unknown parameter: {param}')

            deps = set(self._deps_names.get(param, ()))
            missing_deps = deps.difference(params)
            if missing_deps:
                raise DependencyError(param, missing_deps)

            inffs = set(self._infs_names.get(param, ()))
            missing_inffs = inffs.difference(params)
            if missing_inffs:
                raise InferenceError(param, missing_inffs)

    @classmethod
    def _from_dict(cls, ser: InterDependencies_Dict) -> 'InterDependencies_':
        """
        Construct an InterDependencies_ object from a dictionary
        representation of such an object
        """
        params = ser['parameters']
        deps = cls._extract_deps_from_dict(ser)

        inffs = cls._extract_inffs_from_dict(ser)

        stdls = tuple(ParamSpecBase._from_dict(params[ps_id]) for
                      ps_id in ser['standalones'])

        return cls(dependencies=deps, inferences=inffs, standalones=stdls)

    @classmethod
    def _extract_inffs_from_dict(cls,
                                 ser: InterDependencies_Dict) -> ParamSpecTree:
        params = ser['parameters']
        inffs = {}
        for key, value in ser['inferences'].items():
            inffs_key = ParamSpecBase._from_dict(params[key])
            inffs_vals = tuple(ParamSpecBase._from_dict(params[val]) for
                               val in value)
            inffs.update({inffs_key: inffs_vals})
        return inffs

    @classmethod
    def _extract_deps_from_dict(cls,
                                ser: InterDependencies_Dict) -> ParamSpecTree:
        params = ser['parameters']
        deps = {}
        for key, value in ser['dependencies'].items():
            deps_key = ParamSpecBase._from_dict(params[key])
            deps_vals = tuple(ParamSpecBase._from_dict(params[val]) for
                              val in value)
            deps.update({deps_key: deps_vals})
        return deps

    def __repr__(self) -> str:
        rep = (f"InterDependencies_(dependencies={self.dependencies}, "
               f"inferences={self.inferences}, "
               f"standalones={self.standalones})")
        return rep

    def __eq__(self, other: Any) -> bool:

        def sorter(inp: Iterable['ParamSpecBase']) -> List['ParamSpecBase']:
            return sorted(inp, key=lambda ps: ps.name)

        if not isinstance(other, InterDependencies_):
            return False

        for dic in ['dependencies', 'inferences']:
            our_keys = sorter(getattr(self, dic).keys())
            their_keys = sorter(getattr(other, dic).keys())
            if our_keys != their_keys:
                return False
            for key in our_keys:
                our_values = sorter(getattr(self, dic)[key])
                their_values = sorter(getattr(other, dic)[key])
                if our_values != their_values:
                    return False

        if not self.standalones == other.standalones:
            return False

        return True

    def __contains__(self, ps: ParamSpecBase) -> bool:
        return ps in self._paramspec_to_id

    def __getitem__(self, name: str) -> ParamSpecBase:
        return self._id_to_paramspec[name]
